#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from abc import ABCMeta, abstractmethod

from pyrdp.layer import MCSLayer
from pyrdp.layer.layer import LayerChainItem
from pyrdp.pdu import MCSSendDataIndicationPDU, MCSSendDataRequestPDU


class MCSChannel(LayerChainItem, metaclass=ABCMeta):
    """
    Base class for MCS channels
    A new MCS channel is actually created for every (userID, channelID) pair
    """

    def __init__(self, mcs: MCSLayer, userID: int, channelID: int):
        """
        :param mcs: the MCS layer
        :param userID: the user ID for this channel
        :param channelID: the channel ID for this channel
        """
        super().__init__()
        self.mcs = mcs
        self.userID = userID
        self.channelID = channelID

    def recv(self, data: bytes):
        if self.next is not None:
            self.next.recv(data)


class MCSClientChannel(MCSChannel):
    """
    MCSChannel class and layer for clients.
    Sends SendDataRequest PDUs when sendBytes is called.
    """

    def sendBytes(self, data: bytes):
        pdu = MCSSendDataRequestPDU(self.userID, self.channelID, 0x70, data)
        self.mcs.sendPDU(pdu)


class MCSServerChannel(MCSChannel):
    """
    MCSChannel class and layer for servers.
    Sends SendDataIndication PDUs when sendBytes is called.
    """

    def sendBytes(self, data: bytes):
        pdu = MCSSendDataIndicationPDU(self.userID, self.channelID, 0x70, data)
        self.mcs.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
# flake8: noqa

from pyrdp.pdu.gcc import GCCConferenceCreateRequestPDU, GCCConferenceCreateResponsePDU, GCCPDU
from pyrdp.pdu.mcs import MCSAttachUserConfirmPDU, MCSAttachUserRequestPDU, MCSChannelJoinConfirmPDU, \
    MCSChannelJoinRequestPDU, MCSConnectInitialPDU, MCSConnectResponsePDU, MCSDisconnectProviderUltimatumPDU, \
    MCSDomainParams, MCSErectDomainRequestPDU, MCSPDU, MCSSendDataIndicationPDU, MCSSendDataRequestPDU
from pyrdp.pdu.pdu import PDU
from pyrdp.pdu.player import Color, PlayerBitmapPDU, PlayerConnectionClosePDU, PlayerDeviceMappingPDU, \
    PlayerDirectoryListingRequestPDU, PlayerDirectoryListingResponsePDU, PlayerFileDescription, \
    PlayerFileDownloadCompletePDU, PlayerFileDownloadRequestPDU, PlayerFileDownloadResponsePDU, \
    PlayerForwardingStatePDU, PlayerKeyboardPDU, PlayerMouseButtonPDU, PlayerMouseMovePDU, PlayerMouseWheelPDU, \
    PlayerPDU, PlayerTextPDU
from pyrdp.pdu.rdp.bitmap import BitmapUpdateData
from pyrdp.pdu.rdp.capability import BitmapCacheHostSupportCapability, BitmapCacheV1Capability, BitmapCacheV2Capability, \
    BitmapCapability, BitmapCodec, BitmapCodecsCapability, BrushCapability, Capability, ClientCapsContainer, \
    ColorCacheCapability, ControlCapability, DesktopCompositionCapability, DrawGDIPlusCapability, \
    DrawNineGridCacheCapability, FontCapability, FrameAcknowledgeCapability, GeneralCapability, GlyphCacheCapability, \
    InputCapability, LargePointerCapability, MultifragmentUpdateCapability, NSCodec, OffscreenBitmapCacheCapability, \
    OrderCapability, PointerCapability, RemoteProgramsCapability, RFXCapset, RFXIcap, ServerCapsContainer, \
    ShareCapability, SoundCapability, SurfaceCommandsCapability, VirtualChannelCapability, WindowListCapability, \
    WindowsActivationCapability
from pyrdp.pdu.rdp.client_info import ClientExtraInfo, ClientInfoPDU
from pyrdp.pdu.rdp.connection import ClientChannelDefinition, ClientClusterData, ClientCoreData, ClientDataPDU, \
    ClientNetworkData, ClientSecurityData, ProprietaryCertificate, ServerCertificate, ServerCoreData, ServerDataPDU, \
    ServerNetworkData, ServerSecurityData
from pyrdp.pdu.rdp.fastpath import FastPathBitmapEvent, FastPathEvent, FastPathEventRaw, FastPathInputEvent, \
    FastPathMouseEvent, FastPathOrdersEvent, FastPathOutputEvent, FastPathOutputEvent, FastPathPDU, \
    FastPathScanCodeEvent, FastPathUnicodeEvent
from pyrdp.pdu.rdp.input import ExtendedMouseEvent, KeyboardEvent, MouseEvent, SlowPathInput, SynchronizeEvent, \
    UnicodeKeyboardEvent, UnusedEvent
from pyrdp.pdu.rdp.licensing import LicenseBinaryBlob, LicenseErrorAlertPDU, LicensingPDU
from pyrdp.pdu.rdp.negotiation import NegotiationFailurePDU, NegotiationRequestPDU, NegotiationResponsePDU
from pyrdp.pdu.rdp.ntlmssp import NTLMSSPTSRequestPDU, NTLMSSPPDU, NTLMSSPNegotiatePDU, NTLMSSPChallengePDU, \
    NTLMSSPAuthenticatePDU, NTLMSSPChallengePayloadPDU
from pyrdp.pdu.rdp.pointer import Point, PointerCacheEvent, PointerColorEvent, PointerEvent, PointerNewEvent, \
    PointerPositionEvent, PointerSystemEvent
from pyrdp.pdu.rdp.security import SecurityExchangePDU, SecurityPDU
from pyrdp.pdu.rdp.slowpath import ConfirmActivePDU, ControlPDU, DemandActivePDU, InputPDU, PlaySoundPDU, PointerPDU, \
    SetErrorInfoPDU, ShareControlHeader, ShareDataHeader, SlowPathPDU, SlowPathUnparsedPDU, SuppressOutputPDU, \
    SynchronizePDU, UpdatePDU, PersistentCacheKeysPDU
from pyrdp.pdu.rdp.virtual_channel.clipboard import ClipboardPDU, FormatDataRequestPDU, FormatDataResponsePDU, \
    FormatListPDU, FormatListResponsePDU, FormatName, LongFormatName, ServerMonitorReadyPDU, ShortFormatName, \
    FileContentsRequestPDU, FileContentsResponsePDU
from pyrdp.pdu.rdp.virtual_channel.device_redirection import DeviceAnnounce, DeviceCloseRequestPDU, \
    DeviceCloseResponsePDU, DeviceCreateRequestPDU, DeviceCreateResponsePDU, DeviceDirectoryControlResponsePDU, \
    DeviceDirectoryControlResponsePDU, DeviceIORequestPDU, DeviceIOResponsePDU, DeviceListAnnounceRequest, \
    DeviceQueryDirectoryRequestPDU, DeviceQueryDirectoryResponsePDU, DeviceReadRequestPDU, DeviceReadResponsePDU, \
    DeviceRedirectionCapabilitiesPDU, DeviceRedirectionCapability, DeviceRedirectionClientCapabilitiesPDU, \
    DeviceRedirectionGeneralCapability, DeviceRedirectionPDU, DeviceRedirectionServerCapabilitiesPDU, \
    FileBothDirectoryInformation, FileDirectoryInformation, FileFullDirectoryInformation, FileInformationBase, \
    FileNamesInformation
from pyrdp.pdu.rdp.virtual_channel.virtual_channel import VirtualChannelPDU
from pyrdp.pdu.segmentation import SegmentationPDU
from pyrdp.pdu.tpkt import TPKTPDU
from pyrdp.pdu.x224 import X224ConnectionConfirmPDU, X224ConnectionRequestPDU, X224DataPDU, X224DisconnectRequestPDU, \
    X224ErrorPDU, X224PDU



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from enum import IntEnum


class SegmentationPDUType(IntEnum):
    FAST_PATH = 0
    TPKT = 3
    MASK = 3


#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import asyncio
import logging
from typing import Callable

from PySide6.QtCore import QThread

from pyrdp.logging import LOGGER_NAMES


class LiveThread(QThread):
    """
    Thread for receiving live connection data.
    """

    def __init__(self, host: str, port: int, protocolFactory: Callable[[], asyncio.Protocol]):
        """
        :param host: host to bind to.
        :param port: port to listen on.
        :param protocolFactory: asyncio protocol factory.
        """
        super().__init__()
        self.host = host
        self.port = port
        self.protocolFactory = protocolFactory
        self.loop = asyncio.new_event_loop()

    def run(self):
        asyncio.set_event_loop(self.loop)

        server = self.loop.create_server(self.protocolFactory, host=self.host, port=self.port)
        self.loop.run_until_complete(server)

        logging.getLogger(LOGGER_NAMES.PLAYER).info("Listening for connections on %(listenHost)s:%(listenPort)d", {"listenHost": self.host, "listenPort": self.port})
        self.loop.run_forever()

    def stop(self):
        self.loop.call_soon_threadsafe(self.stopLoop)

    def stopLoop(self):
        self.loop.stop()



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BufferedIOBase
from pathlib import Path
from typing import Dict, List, Optional, Union

from pyrdp.enum import ParserMode, PlayerPDUType
from pyrdp.layer import LayerChainItem, PlayerLayer
from pyrdp.logging import log
from pyrdp.parser import BasicFastPathParser, ClientConnectionParser, \
    ClientInfoParser, ClipboardParser, Parser, SlowPathParser
from pyrdp.pdu import PDU


class Recorder:
    """
    Class that manages recording of RDP events using the provided
    transport layers. Those transport layers are the ones that will
    receive binary data and handle them as they wish. They perform the
    actual recording.
    """

    def __init__(self, transports: List[LayerChainItem]):
        self.parsers: Dict[PlayerPDUType, Parser] = {
            PlayerPDUType.FAST_PATH_INPUT: BasicFastPathParser(ParserMode.CLIENT),
            PlayerPDUType.FAST_PATH_OUTPUT: BasicFastPathParser(ParserMode.SERVER),
            PlayerPDUType.CLIENT_INFO: ClientInfoParser(),
            PlayerPDUType.SLOW_PATH_PDU: SlowPathParser(),
            PlayerPDUType.CLIPBOARD_DATA: ClipboardParser(),
            PlayerPDUType.CLIENT_DATA: ClientConnectionParser(),
        }

        self.topLayers = []
        self.openFileLayers = []
        self.recordFilename = None

        for transport in transports:
            self.addTransport(transport)

    def setRecordFilename(self, filename: str):
        """
        Sets the filename used for the session recording.
        :param filename: the filename
        """
        self.recordFilename = filename

    def addTransport(self, transportLayer: LayerChainItem):
        player = PlayerLayer()
        player.setPrevious(transportLayer)
        self.topLayers.append(player)

        if isinstance(transportLayer, FileLayer):
            self.openFileLayers.append(transportLayer)

    def setParser(self, messageType: PlayerPDUType, parser: Parser):
        """
        Set the parser to use for a given message type.
        """
        self.parsers[messageType] = parser

    def record(self, pdu: Optional[PDU], messageType: PlayerPDUType):
        """
        Encapsulate the pdu properly, then record the data
        """
        if messageType not in self.parsers:
            for layer in self.topLayers:
                layer.sendPDU(pdu)

            return

        if pdu:
            data = self.parsers[messageType].write(pdu)
        else:
            data = b""

        timeStamp = self.getCurrentTimeStamp()

        for layer in self.topLayers:
            layer.sendMessage(data, messageType, timeStamp)

    def getCurrentTimeStamp(self) -> int:
        return PlayerLayer.timeStampFunction()

    def finalize(self):
        """When the recording is finished"""
        for layer in self.openFileLayers:
            layer.close()


class FileLayer(LayerChainItem):
    """
    Layer that saves RDP events to a file for later replay.
    """

    FLUSH_THRESHOLD = 18

    def __init__(self, fileName: Union[str, Path]):
        super().__init__()

        # Buffer data until we have enough bytes for a meaningful replay.
        self.pending = b''

        self.filename = fileName
        self.filename = str(self.filename).replace(":", "_")
        
        self.fd: BufferedIOBase = None

    def sendBytes(self, data: bytes):
        """
        Save data to the file.
        :param data: data to write.
        """

        if not self.fd:
            self.pending += data
            if len(self.pending) > FileLayer.FLUSH_THRESHOLD:
                self.fd = open(str(self.filename), "wb")
                self.fd.write(self.pending)
                self.pending = b''
        elif not self.fd.closed:
            self.fd.write(data)
        else:
            log.error("Recording file handle closed, cannot write message: %(message)s", {"message": data})

    def close(self):
        if self.fd:
            self.fd.close()


#
# Copyright (c) 2014-2015 Sylvain Peyrefitte
# Copyright (c) 2018-2023 GoSecure Inc.
#
# This file is part of PyRDP.
# This file was part of rdpy.
#
# PyRDP is licensed under the GPLv3 or later.
#
# rdpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""
Qt specific code

QRemoteDesktop is a widget use for render in rdpy
"""

import rle
from io import BytesIO

from PySide6.QtCore import QEvent, QPoint, Qt, Signal
from PySide6.QtGui import QColor, QImage, QTransform, QPainter
from PySide6.QtWidgets import QWidget

from pyrdp.logging import log
from pyrdp.player.ImageHandler import ImageHandler


def RDPBitmapToQtImage(width: int, height: int, bitsPerPixel: int, isCompressed: bool, data: bytes):
    """
    Bitmap transformation to Qt object
    :param width: width of bitmap
    :param height: height of bitmap
    :param bitsPerPixel: number of bit per pixel
    :param isCompressed: use RLE compression
    :param data: bitmap data
    """
    image = None
    buf = None

    if bitsPerPixel == 15:
        if isCompressed:
            buf = rle.bitmap_decompress(data, width, height, 2)
            image = QImage(buf, width, height, QImage.Format_RGB555)
        else:
            buf = data
            image = QImage(buf, width, height, QImage.Format_RGB555).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))

    elif bitsPerPixel == 16:
        if isCompressed:
            buf = rle.bitmap_decompress(data, width, height, 2)
            image = QImage(buf, width, height, QImage.Format_RGB16)
        else:
            buf = data
            image = QImage(buf, width, height, QImage.Format_RGB16).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))

    elif bitsPerPixel == 24:
        if isCompressed:
            buf = rle.bitmap_decompress(data, width, height, 3)

            # This is a ugly patch because there is a bug in the 24bpp decompression in rle.c
            # where the red and the blue colors are inverted. Fixing this in python causes a performance
            # issue, but at least it shows the good colors.
            buf2 = BytesIO(buf)
            while buf2.tell() < len(buf2.getvalue()):
                pixel = buf2.read(3)
                buf[buf2.tell() - 3] = pixel[2]
                buf[buf2.tell() - 1] = pixel[0]

            image = QImage(buf, width, height, QImage.Format_RGB888)
        else:
            buf = data
            image = QImage(buf, width, height, QImage.Format_RGB888).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))

    elif bitsPerPixel == 32:
        if isCompressed:
            buf = rle.bitmap_decompress(data, width, height, 4)
            image = QImage(buf, width, height, QImage.Format_RGB32)
        else:
            buf = data
            image = QImage(buf, width, height, QImage.Format_RGB32).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
    elif bitsPerPixel == 8:
        if isCompressed:
            _buf = rle.bitmap_decompress(data, width, height, 1)
            buf = convert8bppTo16bpp(_buf)
            image = QImage(buf, width, height, QImage.Format_RGB16)
        else:
            buf = convert8bppTo16bpp(data)
            image = QImage(buf, width, height, QImage.Format_RGB16).transformed(QTransform(1.0, 0.0, 0.0, -1.0, 0.0, 0.0))
    else:
        log.error("Receive image in bad format")
        image = QImage(width, height, QImage.Format_RGB32)
    return (image, buf)


def convert8bppTo16bpp(buf: bytes):
    r"""
    WARNING: The actual 8bpp images work by using a color palette, which this method does not use.
    This method instead tries to transform indices into colors. This results in a weird looking image,
    but it can still be useful to see whats happening ¯\_(ツ)_/¯
    """
    buf2 = bytearray(len(buf) * 2)
    i = 0
    for pixel in buf:
        r = (pixel & 0b11000000) >> 6
        g = (pixel & 0b00111000) >> 3
        b = (pixel & 0b00000111) >> 0
        buf2[i] = (b << 3)
        buf2[i + 1] = (g << 0) | (r << 5)
        i += 2

    return buf2


class QRemoteDesktop(QWidget, ImageHandler):
    """
    Qt RDP display widget. It is the widget directly responsible to display the "screen" of the
    client in the RDP session being shown/replayed.
    """

    # This signal can be used by other objects to run code on the main thread. The argument is a callable.
    mainThreadHook = Signal(object)

    def __init__(self, width: int, height: int, parent: QWidget = None):
        """
        :param width: width of widget
        :param height: height of widget
        :param parent: parent widget
        """
        super().__init__(parent)

        self.ratio = 1
        """ Scale factor used to render the RDP session on the player."""

        self.sessionWidth = width
        self.sessionHeight = height

        self.scaleToWindow = False

        # Buffer image
        self._buffer: QImage = QImage(width, height, QImage.Format_ARGB32_Premultiplied)

        # Set correct size
        self.resize(width, height)
        # Bind mouse event
        self.setMouseTracking(True)
        self.mouseX = width // 2
        self.mouseY = height // 2

        self.mainThreadHook.connect(self.runOnMainThread)

    def notifyImage(self, x: int, y: int, qimage: QImage, width: int, height: int):
        """
        Draw an image on the buffer.
        :param x: x position of the new image
        :param y: y position of the new image
        :param qimage: new QImage
        :param width: width of the new image
        :param height: height of the new image
        """

        # Fill buffer image
        qp = QPainter(self._buffer)
        qp.drawImage(x, y, qimage, 0, 0, width, height)

        # Force update
        self.update()

    def resize(self, width: int, height: int):
        """
        Resize the image buffer. This is called when the clientData is parsed, which
        contains the screen size used for the connection.
        :param width: new width of the replay client's screen
        :param height: new height of the replay client's screen.
        """
        self._buffer = QImage(width, height, QImage.Format_ARGB32_Premultiplied)
        self.sessionWidth = width
        self.sessionHeight = height
        self._updateWidgetSize()

    def update(self):
        QWidget.update(self)

    @property
    def screen(self):
        return self._buffer

    def runOnMainThread(self, target: callable):
        target()

    def setMousePosition(self, x: int, y: int):
        self.mouseX = x
        self.mouseY = y
        self.update()

    def scale(self, scale):
        """
        Rescale the current widget to a percentage of the height of the RDP session.
        :param scale: Ex: 0.5 for 50% height and 50% width.
        """
        self.ratio = scale
        self._updateWidgetSize()

    def _updateWidgetSize(self):
        """
        Size the widget according to if we are scaled or not
        """
        if self.scaleToWindow:
            # resize widget to parent size: will hide scrollbars
            super().resize(self.parent().size())
        else:
            # resize widget to session size: will show scrollbars if required
            super().resize(self.sessionWidth, self.sessionHeight)

    def setScaleToWindow(self, status):
        self.scaleToWindow = status > 0

    def paintEvent(self, e: QEvent):
        """
        Call when Qt renderer engine estimate that is needed
        :param e: the event
        """
        ratio = self.ratio if self.scaleToWindow else 1
        qp = QPainter(self)
        qp.drawImage(0, 0, self._buffer.scaled(self.sessionWidth * ratio, self.sessionHeight * ratio, aspectMode=Qt.KeepAspectRatio))
        qp.setBrush(QColor.fromRgb(255, 255, 0, 180))
        qp.drawEllipse(QPoint(self.mouseX * ratio, self.mouseY * ratio), 5, 5)

    def clear(self):
        self._buffer = QImage(self._buffer.width(), self._buffer.height(), QImage.Format_ARGB32_Premultiplied)
        self.setMousePosition(self._buffer.width() // 2, self._buffer.height() // 2)
        self.repaint()



#
# This file is part of the PyRDP project.
# Copyright (C) 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import LoggerAdapter
from Crypto.PublicKey.RSA import RsaKey

from pyrdp.core import decodeUTF16LE
from pyrdp.enum import ClientInfoFlags, PlayerPDUType
from pyrdp.layer import SecurityLayer
from pyrdp.mitm.config import MITMConfig
from pyrdp.mitm.state import RDPMITMState
from pyrdp.parser import ClientInfoParser
from pyrdp.pdu import SecurityExchangePDU
from pyrdp.recording import Recorder
from pyrdp.security.crypto import RSA


class SecurityMITM:
    """
    MITM component for the security layer.
    """

    def __init__(self, client: SecurityLayer, server: SecurityLayer, log: LoggerAdapter, config: MITMConfig, state: RDPMITMState, recorder: Recorder):
        """
        :param client: security layer for the client side
        :param server: security layer for the server side
        :param log: logger for this component
        :param config: the MITM configuration
        :param state: the MITM state
        :param recorder: recorder for this connection
        """
        self.client = client
        self.server = server
        self.log = log
        self.state = state
        self.config = config
        self.recorder = recorder

        self.client.createObserver(
            onLicensingDataReceived = self.onClientLicensingData,
            onSecurityExchangeReceived = self.onSecurityExchange,
            onClientInfoReceived=self.onClientInfo,
        )

        self.server.createObserver(
            onLicensingDataReceived = self.onServerLicensingData,
        )

    def onSecurityExchange(self, pdu: SecurityExchangePDU):
        """
        Set the security settings' client random from the security exchange.
        :param pdu: the security exchange
        """
        clientRandom = RSA(self.state.rc4RSAKey).decrypt(pdu.clientRandom[:: -1])[:: -1]
        self.state.securitySettings.setClientRandom(clientRandom)

        self.server.sendSecurityExchange(self.state.securitySettings.encryptClientRandom())

    def onClientInfo(self, data: bytes):
        """
        Log the client connection information and replace the username and password if applicable.
        :param data: the client info data
        """
        pdu = ClientInfoParser().parse(data)

        clientAddress = None

        if pdu.extraInfo:
            clientAddress = decodeUTF16LE(pdu.extraInfo.clientAddress)

        self.log.info("Client Info: username = %(username)r, password = %(password)r, domain = %(domain)r, clientAddress = %(clientAddress)r", {
            "username": pdu.username,
            "password": pdu.password,
            "domain": pdu.domain,
            "clientAddress": clientAddress
        })

        self.recorder.record(pdu, PlayerPDUType.CLIENT_INFO)

        # If set, replace the provided username and password to connect the user regardless of
        # the credentials they entered.
        if self.config.replacementUsername is not None:
            pdu.username = self.config.replacementUsername
        if self.config.replacementPassword is not None:
            pdu.password = self.config.replacementPassword

        if self.config.replacementUsername is not None and self.config.replacementPassword is not None:
            pdu.flags |= ClientInfoFlags.INFO_AUTOLOGON

        # Tell the server we don't want compression (unsure of the effectiveness of these flags)
        pdu.flags &= ~ClientInfoFlags.INFO_COMPRESSION
        pdu.flags &= ~ClientInfoFlags.INFO_CompressionTypeMask

        self.log.debug("Sending %(pdu)s", {"pdu": pdu})
        self.server.sendClientInfo(pdu)

    def onServerLicensingData(self, data: bytes):
        """
        Forward licensing data to the client and disable security headers if TLS is in use.
        :param data: the licensing data
        """
        if self.state.useTLS:
            self.client.securityHeaderExpected = False
            self.server.securityHeaderExpected = False

        self.client.sendLicensing(data)

    def onClientLicensingData(self, data: bytes):
        """
        Forward licensing data to the server and disable security headers if TLS is in use.
        :param data: the licensing data
        """
        if self.state.useTLS:
            self.client.securityHeaderExpected = False
            self.server.securityHeaderExpected = False

        self.server.sendLicensing(data)


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import GCCPDUType
from pyrdp.pdu.pdu import PDU


class GCCPDU(PDU):
    """
    Base PDU class for GCC (T.124) PDUs. Every GCC PDU has a header and a payload.
    """

    def __init__(self, header: GCCPDUType, payload: bytes):
        """
        :param header: GCC PDU type.
        :param payload: GCC's payload (so probably some RDP connection data).
        """

        PDU.__init__(self, payload)
        self.header = header


class GCCConferenceCreateRequestPDU(GCCPDU):
    def __init__(self, conferenceName: str, payload: bytes):
        """
        :param conferenceName: the GCC conference name.
        :param payload: GCC's payload (so probably some RDP connection data).
        """
        GCCPDU.__init__(self, GCCPDUType.CREATE_CONFERENCE_REQUEST, payload)
        self.conferenceName = conferenceName


class GCCConferenceCreateResponsePDU(GCCPDU):
    def __init__(self, nodeID: int, tag: int, result: int, payload: bytes):
        """
        :param nodeID: GCC node ID.
        :param tag: GCC tag.
        :param result: response result.
        :param payload: GCC's payload (so probably some RDP connection data).
        """
        GCCPDU.__init__(self, GCCPDUType.CREATE_CONFERENCE_RESPONSE, payload)
        self.nodeID = nodeID
        self.tag = tag
        self.result = result



#
# This file is part of the PyRDP project.
# Copyright (C) 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO

from pyrdp.enum import NTLMSSPMessageType
from pyrdp.pdu.pdu import PDU


class NTLMSSPPDU(PDU):
    def __init__(self, messageType: NTLMSSPMessageType):
        super().__init__()
        self.messageType = messageType


class NTLMSSPNegotiatePDU(NTLMSSPPDU):
    def __init__(self):
        super().__init__(NTLMSSPMessageType.NEGOTIATE_MESSAGE)


class NTLMSSPChallengePDU(NTLMSSPPDU):
    def __init__(self, serverChallenge: bytes):
        super().__init__(NTLMSSPMessageType.CHALLENGE_MESSAGE)
        self.serverChallenge = serverChallenge


class NTLMSSPChallengePayloadPDU(PDU):
    """
    Payload of CHALLENGE message containing a data array
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
    """
    def __init__(self, workstation: str):
        super().__init__()
        self.workstation = workstation


class NTLMSSPAuthenticatePDU(NTLMSSPPDU):
    def __init__(self, user: str, domain: str, proof: bytes, response: bytes):
        super().__init__(NTLMSSPMessageType.AUTHENTICATE_MESSAGE)
        self.user = user
        self.domain = domain
        self.proof = proof
        self.response = response


class NTLMSSPTSRequestPDU(PDU):
    """
    PDU for TSRequest structures used by CredSSP (client/server) for SPNEGO and Kerberos/NTLM messages
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/6aac4dea-08ef-47a6-8747-22ea7f6d8685
    """
    def __init__(self, version: int, negoTokens: BytesIO):
        super().__init__()
        self.version = version
        self.negoTokens = negoTokens



# This file is part of the PyRDP project.
#
# Copyright (C) 2020-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import BitmapFlags, CapabilityType, SlowPathUpdateType
from pyrdp.parser import BitmapParser, FastPathOutputParser, OrdersParser
from pyrdp.pdu import BitmapUpdateData, FastPathBitmapEvent, FastPathOutputEvent, FastPathOrdersEvent, UpdatePDU
from pyrdp.player import ImageHandler
from pyrdp.player.BaseEventHandler import BaseEventHandler
from pyrdp.player.gdi.draw import GdiQtFrontend
from pyrdp.ui import RDPBitmapToQtImage

import logging


class RenderingEventHandler(BaseEventHandler):
    """Abstract class for rendering handlers."""

    def __init__(self, imageHandler: ImageHandler):
        BaseEventHandler.__init__(self)
        self._fastPath = FastPathOutputParser()
        self._bitmap = BitmapParser()
        self.gdi: GdiQtFrontend = None
        self._orders: OrdersParser = None
        self.log = logging.getLogger(__name__)
        self.imageHandler = imageHandler

    def onCapabilities(self, caps):
        if CapabilityType.CAPSTYPE_ORDER in caps:
            self.gdi = GdiQtFrontend(self.imageHandler)
            self._orders = OrdersParser(self.gdi)
            self._orders.onCapabilities(caps)

    # Generic Parsing Routines.
    def onFastPathOutput(self, event: FastPathOutputEvent):
        self.onBeginRender()
        if isinstance(event, FastPathBitmapEvent):
            parsed = self._fastPath.parseBitmapEvent(event)
            for bmp in parsed.bitmapUpdateData:
                self.onBitmap(bmp)

        elif isinstance(event, FastPathOrdersEvent):
            if self._orders is None:
                self.log.error('Received Unexpected Drawing Orders!')
                return
            self.onBeginRender()
            self._orders.parse(event)

        self.onFinishRender()

    def onSlowPathUpdate(self, pdu: UpdatePDU):
        if pdu.updateType == SlowPathUpdateType.SLOWPATH_UPDATETYPE_BITMAP:
            self.onBeginRender()

            updates = self._bitmap.parseBitmapUpdateData(pdu.updateData)
            for bmp in updates:
                self.onBitmap(bmp)

            self.onFinishRender()

    def onBitmap(self, bitmapData: BitmapUpdateData):
        image, _ = RDPBitmapToQtImage(
            bitmapData.width,
            bitmapData.heigth,
            bitmapData.bitsPerPixel,
            bitmapData.flags & BitmapFlags.BITMAP_COMPRESSION != 0,
            bitmapData.bitmapData
        )

        self.imageHandler.notifyImage(
            bitmapData.destLeft,
            bitmapData.destTop,
            image,
            bitmapData.destRight - bitmapData.destLeft + 1,
            bitmapData.destBottom - bitmapData.destTop + 1)

    def onBeginRender(self):
        pass

    def onFinishRender(self):
        pass



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import CapabilityType
from pyrdp.pdu.pdu import PDU


class Capability(PDU):
    def __init__(self, capabilityType: int, rawData=b""):
        super().__init__()
        self.capabilityType = capabilityType
        self.rawData = rawData


class GeneralCapability(Capability):
    def __init__(self, majorType, minorType, protocolVersion, generalCompressionTypes, extraFlags, updateCapabilityFlag, remoteUnshareFlag, generalCompressionLevel, refreshRectSupport, suppressOutputSupport):
        Capability.__init__(self, CapabilityType.CAPSTYPE_GENERAL)
        self.majorType = majorType
        self.minorType = minorType
        self.protocolVersion = protocolVersion
        self.generalCompressionTypes = generalCompressionTypes
        self.extraFlags = extraFlags
        self.updateCapabilityFlag = updateCapabilityFlag
        self.remoteUnshareFlag = remoteUnshareFlag
        self.generalCompressionLevel = generalCompressionLevel
        self.refreshRectSupport = refreshRectSupport
        self.suppressOutputSupport = suppressOutputSupport


class BitmapCapability(Capability):
    def __init__(self, preferredBitsPerPixel, receive1BitPerPixel, receive4BitsPerPixel, receive8BitsPerPixel, desktopWidth, desktopHeight, desktopResizeFlag, bitmapCompressionFlag, highColorFlags, drawingFlags, multipleRectangleSupport):
        Capability.__init__(self, CapabilityType.CAPSTYPE_BITMAP)
        self.preferredBitsPerPixel = preferredBitsPerPixel
        self.receive1BitPerPixel = receive1BitPerPixel
        self.receive4BitsPerPixel = receive4BitsPerPixel
        self.receive8BitsPerPixel = receive8BitsPerPixel
        self.desktopWidth = desktopWidth
        self.desktopHeight = desktopHeight
        self.desktopResizeFlag = desktopResizeFlag
        self.bitmapCompressionFlag = bitmapCompressionFlag
        self.highColorFlags = highColorFlags
        self.drawingFlags = drawingFlags
        self.multipleRectangleSupport = multipleRectangleSupport


class OrderCapability(Capability):
    def __init__(self, terminalDescriptor, desktopSaveXGranularity, desktopSaveYGranularity, maximumOrderLevel, numberFonts, orderFlags, orderSupport, textFlags, orderSupportExFlags, desktopSaveSize, textANSICodePage):
        Capability.__init__(self, CapabilityType.CAPSTYPE_ORDER)
        self.terminalDescriptor = terminalDescriptor
        self.desktopSaveXGranularity = desktopSaveXGranularity
        self.desktopSaveYGranularity = desktopSaveYGranularity
        self.maximumOrderLevel = maximumOrderLevel
        self.numberFonts = numberFonts
        self.orderFlags = orderFlags
        self.orderSupport = orderSupport
        self.textFlags = textFlags
        self.orderSupportExFlags = orderSupportExFlags
        self.desktopSaveSize = desktopSaveSize
        self.textANSICodePage = textANSICodePage


class BitmapCacheV1Capability(Capability):
    def __init__(self, cache0Entries, cache0MaximumCellSize, cache1Entries, cache1MaximumCellSize, cache2Entries, cache2MaximumCellSize):
        Capability.__init__(self, CapabilityType.CAPSTYPE_BITMAPCACHE)
        self.cache0Entries = cache0Entries
        self.cache0MaximumCellSize = cache0MaximumCellSize
        self.cache1Entries = cache1Entries
        self.cache1MaximumCellSize = cache1MaximumCellSize
        self.cache2Entries = cache2Entries
        self.cache2MaximumCellSize = cache2MaximumCellSize


class BitmapCacheV2Capability(Capability):
    def __init__(self, cacheFlags, numCellCaches, bitmapCache0CellInfo, bitmapCache1CellInfo, bitmapCache2CellInfo, bitmapCache3CellInfo, bitmapCache41CellInfo):
        Capability.__init__(self, CapabilityType.CAPSTYPE_BITMAPCACHE_REV2)
        self.cacheFlags = cacheFlags
        self.numCellCaches = numCellCaches
        self.bitmapCache0CellInfo = bitmapCache0CellInfo
        self.bitmapCache1CellInfo = bitmapCache1CellInfo
        self.bitmapCache2CellInfo = bitmapCache2CellInfo
        self.bitmapCache3CellInfo = bitmapCache3CellInfo
        self.bitmapCache41CellInfo = bitmapCache41CellInfo


class ControlCapability(Capability):
    def __init__(self, controlFlags, remoteDetachFlag, controlInterest, detachInterest):
        Capability.__init__(self, CapabilityType.CAPSTYPE_CONTROL)
        self.controlFlags = controlFlags
        self.remoteDetachFlag = remoteDetachFlag
        self.controlInterest = controlInterest
        self.detachInterest = detachInterest


class WindowsActivationCapability(Capability):
    def __init__(self, helpKeyFlag, helpKeyIndexFlag, helpExtendedKeyFlag, windowManagerKeyFlag):
        Capability.__init__(self, CapabilityType.CAPSTYPE_ACTIVATION)
        self.helpKeyFlag = helpKeyFlag
        self.helpKeyIndexFlag = helpKeyIndexFlag
        self.helpExtendedKeyFlag = helpExtendedKeyFlag
        self.windowManagerKeyFlag = windowManagerKeyFlag


class PointerCapability(Capability):
    def __init__(self, colorPointerFlag, colorPointerCacheSize, pointerCacheSize):
        Capability.__init__(self, CapabilityType.CAPSTYPE_POINTER)
        self.colorPointerFlag = colorPointerFlag
        self.colorPointerCacheSize = colorPointerCacheSize
        self.pointerCacheSize = pointerCacheSize


class ShareCapability(Capability):
    def __init__(self, nodeID):
        Capability.__init__(self, CapabilityType.CAPSTYPE_SHARE)
        self.nodeID = nodeID


class ColorCacheCapability(Capability):
    def __init__(self, colorTableCacheSize):
        Capability.__init__(self, CapabilityType.CAPSTYPE_COLORCACHE)
        self.colorTableCacheSize = colorTableCacheSize


class SoundCapability(Capability):
    def __init__(self, soundFlags):
        Capability.__init__(self, CapabilityType.CAPSTYPE_SOUND)
        self.soundFlags = soundFlags


class InputCapability(Capability):
    def __init__(self, inputFlags, keyboardLayout, keyboardType, keyboardSubType, keyboardFunctionKey, imeFileName):
        Capability.__init__(self, CapabilityType.CAPSTYPE_INPUT)
        self.inputFlags = inputFlags
        self.keyboardLayout = keyboardLayout
        self.keyboardType = keyboardType
        self.keyboardSubType = keyboardSubType
        self.keyboardFunctionKey = keyboardFunctionKey
        self.imeFileName = imeFileName


class FontCapability(Capability):
    def __init__(self, fontSupportFlags):
        Capability.__init__(self, CapabilityType.CAPSTYPE_FONT)
        self.fontSupportFlags = fontSupportFlags


class BrushCapability(Capability):
    def __init__(self, brushSupportLevel):
        Capability.__init__(self, CapabilityType.CAPSTYPE_BRUSH)
        self.brushSupportLevel = brushSupportLevel


class OffscreenBitmapCacheCapability(Capability):
    def __init__(self, offscreenSupportLevel, offscreenCacheSize, offscreenCacheEntries):
        Capability.__init__(self, CapabilityType.CAPSTYPE_OFFSCREENCACHE)
        self.offscreenSupportLevel = offscreenSupportLevel
        self.offscreenCacheSize = offscreenCacheSize
        self.offscreenCacheEntries = offscreenCacheEntries


class BitmapCacheHostSupportCapability(Capability):
    def __init__(self, cacheVersion):
        Capability.__init__(self, CapabilityType.CAPSTYPE_BITMAPCACHE_HOSTSUPPORT)
        self.cacheVersion = cacheVersion


class GlyphCacheCapability(Capability):
    def __init__(self, glyphCache, fragCache, glyphSupportLevel):
        Capability.__init__(self, CapabilityType.CAPSTYPE_GLYPHCACHE)
        self.glyphCache = glyphCache
        self.fragCache = fragCache
        self.glyphSupportLevel = glyphSupportLevel

class VirtualChannelCapability(Capability):
    def __init__(self, flags, vcChunkSize):
        Capability.__init__(self, CapabilityType.CAPSTYPE_VIRTUALCHANNEL)
        self.flags = flags
        self.vcChunkSize = vcChunkSize


class DrawNineGridCacheCapability(Capability):
    def __init__(self, drawNineGridSupportLevel, drawNineGridCacheSize, drawNineGridCacheEntries):
        Capability.__init__(self, CapabilityType.CAPSTYPE_DRAWNINEGRIDCACHE)
        self.drawNineGridSupportLevel = drawNineGridSupportLevel
        self.drawNineGridCacheSize = drawNineGridCacheSize
        self.drawNineGridCacheEntries = drawNineGridCacheEntries


class DrawGDIPlusCapability(Capability):
    def __init__(self, drawGDIPlusSupportLevel, gdipVersion, drawGDIPlusCacheLevel, gdipCacheEntries, gdipCacheChunkSize, gdipImageCacheProperties):
        Capability.__init__(self, CapabilityType.CAPSTYPE_DRAWGDIPLUS)
        self.drawGDIPlusSupportLevel = drawGDIPlusSupportLevel
        self.gdipVersion = gdipVersion
        self.drawGDIPlusCacheLevel = drawGDIPlusCacheLevel
        self.gdipCacheEntries = gdipCacheEntries
        self.gdipCacheChunkSize = gdipCacheChunkSize
        self.gdipImageCacheProperties = gdipImageCacheProperties


class RemoteProgramsCapability(Capability):
    def __init__(self, railSupportLevel):
        Capability.__init__(self, CapabilityType.CAPSTYPE_RAIL)
        self.railSupportLevel = railSupportLevel


class WindowListCapability(Capability):
    def __init__(self, wndSupportLevel, numIconCaches, numIconCacheEntries):
        Capability.__init__(self, CapabilityType.CAPSTYPE_WINDOW)
        self.wndSupportLevel = wndSupportLevel
        self.numIconCaches = numIconCaches
        self.numIconCacheEntries = numIconCacheEntries


class DesktopCompositionCapability(Capability):
    def __init__(self, compDeskSupportLevel):
        Capability.__init__(self, CapabilityType.CAPSETTYPE_COMPDESK)
        self.compDeskSupportLevel = compDeskSupportLevel


class MultifragmentUpdateCapability(Capability):
    def __init__(self, maxRequestSize):
        Capability.__init__(self, CapabilityType.CAPSETTYPE_MULTIFRAGMENTUPDATE)
        self.maxRequestSize = maxRequestSize


class LargePointerCapability(Capability):
    def __init__(self, largePointerSupportFlags):
        Capability.__init__(self, CapabilityType.CAPSETTYPE_LARGE_POINTER)
        self.largePointerSupportFlags = largePointerSupportFlags


class SurfaceCommandsCapability(Capability):
    def __init__(self, cmdFlags, reserved):
        Capability.__init__(self, CapabilityType.CAPSETTYPE_SURFACE_COMMANDS)
        self.cmdFlags = cmdFlags
        self.reserved = reserved


class BitmapCodecsCapability(Capability):
    def __init__(self, supportedBitmapCodecs):
        Capability.__init__(self, CapabilityType.CAPSETTYPE_BITMAP_CODECS)
        self.supportedBitmapCodecs = supportedBitmapCodecs


class BitmapCodec:
    def __init__(self, guid, id, properties):
        self.guid = guid
        self.id = id
        self.properties = properties


class NSCodec:
    def __init__(self, allowDynamicFidelity, allowSubsampling, colorLossLevel):
        self.allowDynamicFidelity = allowDynamicFidelity
        self.allowSubsampling = allowSubsampling
        self.colorLossLevel = colorLossLevel


class ClientCapsContainer:
    def __init__(self, captureFlags, rfxCapset):
        self.captureFlags = captureFlags
        self.rfxCaps = rfxCapset


class RFXCapset:
    def __init__(self, id, capsetType, icaps):
        self.id = id
        self.capsetType = capsetType
        self.icaps = icaps


class RFXIcap:
    def __init__(self, flags, entropyBits):
        self.flags = flags
        self.entropyBits = entropyBits
        

class ServerCapsContainer:
    def __init__(self, reserved):
        self.reserved = reserved


class FrameAcknowledgeCapability(Capability):
    def __init__(self, maxUnacknowledgedFrameCount):
        Capability.__init__(self, CapabilityType.CAPSSETTYPE_FRAME_ACKNOWLEDGE)
        self.maxUnacknowledgedFrameCount = maxUnacknowledgedFrameCount



#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#
from configparser import ConfigParser, ExtendedInterpolation

import appdirs
CONFIG_DIR = appdirs.user_config_dir("pyrdp", "pyrdp")


def load(path: str, fallback: ConfigParser = None) -> ConfigParser:
    """
    Retrieve the PyRDP settings from a file

    :param path: The path of the file to load.
    :param fallback: The fallback configuration path.

    :returns: A ConfigParser instance with the loaded settings.
    :throws Exception: When the fallback configuration is missing and no configuration is found.
    """
    config = ConfigParser(interpolation=ExtendedInterpolation())
    config.optionxform = str
    try:
        if len(config.read(path)) > 0:
            return config
    except Exception:
        # Fallback to default settings.
        pass

    if not fallback:
        raise Exception('Invalid configuration with no fallback specified')
    return fallback



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from logging import LoggerAdapter
from typing import Callable, Dict

from pyrdp.enum import ClientCapabilityFlag, EncryptionLevel, EncryptionMethod, HighColorDepth, MCSChannelName, \
    PlayerPDUType, SupportedColorDepth
from pyrdp.layer import MCSLayer
from pyrdp.logging.StatCounter import StatCounter, STAT
from pyrdp.mcs import MCSClientChannel, MCSServerChannel
from pyrdp.mitm.state import RDPMITMState
from pyrdp.parser import ClientConnectionParser, GCCParser, ServerConnectionParser
from pyrdp.pdu import GCCConferenceCreateRequestPDU, GCCConferenceCreateResponsePDU, MCSAttachUserConfirmPDU, \
    MCSAttachUserRequestPDU, MCSChannelJoinConfirmPDU, MCSChannelJoinRequestPDU, MCSConnectInitialPDU, \
    MCSConnectResponsePDU, MCSDisconnectProviderUltimatumPDU, MCSErectDomainRequestPDU, MCSSendDataIndicationPDU, \
    MCSSendDataRequestPDU, ProprietaryCertificate, ServerDataPDU, ServerSecurityData, ClientDataPDU
from pyrdp.recording import Recorder


class MCSMITM:
    """
    This is the MITM component for the MCS layer.
    This component removes a number of flags in the connection sequence to avoid using RDP features that are not
    implemented. It also takes care of the RDP channel map and initializes the security settings. It relies on an
    external callback for building MCS channels when a join request is accepted.
    """

    def __init__(self, client: MCSLayer, server: MCSLayer, state: RDPMITMState, recorder: Recorder,
                 buildChannelCallback: Callable[[MCSServerChannel, MCSClientChannel], None],
                 log: LoggerAdapter, statCounter: StatCounter):
        """
        :param client: MCS layer for the client side
        :param server: MCS layer for the server side
        :param state: the RDP MITM shared state
        :param recorder: the recorder for this session
        :param buildChannelCallback: function called when MCS channels are built
        :param log: logger for the MCS layer.
        """

        self.log = log
        self.statCounter = statCounter
        self.client = client
        self.server = server
        self.state = state
        self.recorder = recorder
        self.buildChannelCallback = buildChannelCallback

        self.clientChannels: Dict[int, MCSServerChannel] = {}
        """MCS channels for the client side. From the point of view of the MITM, these are server MCS channels."""

        self.serverChannels: Dict[int, MCSClientChannel] = {}
        """MCS channels for the server side. From the point of view of the MITM, these are client MCS channels."""

        self.client.createObserver(
            onConnectInitial = self.onConnectInitial,
            onErectDomainRequest = self.onErectDomainRequest,
            onAttachUserRequest = self.onAttachUserRequest,
            onChannelJoinRequest = self.onChannelJoinRequest,
            onSendDataRequest = self.onSendDataRequest,
            onDisconnectProviderUltimatum = self.onClientDisconnectProviderUltimatum
        )

        self.server.createObserver(
            onConnectResponse = self.onConnectResponse,
            onAttachUserConfirm = self.onAttachUserConfirm,
            onChannelJoinConfirm = self.onChannelJoinConfirm,
            onSendDataIndication = self.onSendDataIndication,
            onDisconnectProviderUltimatum = self.onServerDisconnectProviderUltimatum
        )

    def onConnectInitial(self, pdu: MCSConnectInitialPDU):
        """
        Parse client connection information, change some settings, disable some unimplemented features, and record
        the client data.
        :param pdu: the connect initial PDU
        """

        gccParser = GCCParser()
        rdpClientConnectionParser = ClientConnectionParser()
        gccConferenceCreateRequestPDU: GCCConferenceCreateRequestPDU = gccParser.parse(pdu.payload)
        rdpClientDataPDU: ClientDataPDU = rdpClientConnectionParser.parse(gccConferenceCreateRequestPDU.payload)

        # FIPS is not implemented, so remove this flag if it's set
        rdpClientDataPDU.securityData.encryptionMethods &= ~EncryptionMethod.ENCRYPTION_FIPS
        rdpClientDataPDU.securityData.extEncryptionMethods &= ~EncryptionMethod.ENCRYPTION_FIPS

        self.state.windowSize = (rdpClientDataPDU.coreData.desktopWidth, rdpClientDataPDU.coreData.desktopHeight)

        if self.state.config.downgrade:
            #  This disables the support for the Graphics pipeline extension, which is a completely different way to
            #  transfer graphics from server to client. https://msdn.microsoft.com/en-us/library/dn366933.aspx
            rdpClientDataPDU.coreData.earlyCapabilityFlags &= ~ClientCapabilityFlag.RNS_UD_CS_SUPPORT_DYNVC_GFX_PROTOCOL

            #  Remove 24bpp and 32bpp support, fall back to 16bpp.
            #  2018-12-14: This is only there because there is a bug in the pyrdp player where 24bpp
            #  decompression in rle.c causes random crashes. If this bug is fixed, we could remove this.
            rdpClientDataPDU.coreData.supportedColorDepths &= ~SupportedColorDepth.RNS_UD_32BPP_SUPPORT
            rdpClientDataPDU.coreData.supportedColorDepths &= ~SupportedColorDepth.RNS_UD_24BPP_SUPPORT
            rdpClientDataPDU.coreData.highColorDepth &= ~HighColorDepth.HIGH_COLOR_24BPP

            if rdpClientDataPDU.coreData.highColorDepth == 0:
                # Means the requested color depth was 24bpp, fallback to 16bpp
                rdpClientDataPDU.coreData.highColorDepth |= HighColorDepth.HIGH_COLOR_16BPP

            rdpClientDataPDU.coreData.earlyCapabilityFlags &= ~ClientCapabilityFlag.RNS_UD_CS_WANT_32BPP_SESSION

        self.recorder.record(rdpClientDataPDU, PlayerPDUType.CLIENT_DATA)

        if rdpClientDataPDU.networkData:
            self.state.channelDefinitions = rdpClientDataPDU.networkData.channelDefinitions
            if "MS_T120" in map(lambda channelDef: channelDef.name, rdpClientDataPDU.networkData.channelDefinitions):
                self.log.info("Bluekeep (CVE-2019-0708) scan or exploit attempt detected.", {"bluekeep": True})

        serverGCCPDU = GCCConferenceCreateRequestPDU("1", rdpClientConnectionParser.write(rdpClientDataPDU))
        serverMCSPDU = MCSConnectInitialPDU(
            pdu.callingDomain,
            pdu.calledDomain,
            pdu.upward,
            pdu.targetParams,
            pdu.minParams,
            pdu.maxParams,
            gccParser.write(serverGCCPDU)
        )

        self.log.info("Client hostname %(clientName)s", {"clientName": rdpClientDataPDU.coreData.clientName.strip("\x00")})

        self.server.sendPDU(serverMCSPDU)

    def onConnectResponse(self, pdu: MCSConnectResponsePDU):
        """
        Parse server connection information. Initialize security settings and map channel IDs to channel names.
        :param pdu: the connect response PDU
        """

        if pdu.result != 0:
            self.client.sendPDU(pdu)
        else:
            # Parse response PDUs
            gccParser = GCCParser()
            rdpParser = ServerConnectionParser()
            gccPDU: GCCConferenceCreateResponsePDU = gccParser.parse(pdu.payload)
            serverData = rdpParser.parse(gccPDU.payload)

            # Save security settings
            self.state.securitySettings.setEncryptionMethod(serverData.securityData.encryptionMethod)
            self.state.securitySettings.setServerRandom(serverData.securityData.serverRandom)

            if serverData.securityData.serverCertificate:
                self.state.securitySettings.setServerPublicKey(serverData.securityData.serverCertificate.publicKey)

            # Map channel names to IDs
            self.state.channelMap[serverData.networkData.mcsChannelID] = MCSChannelName.IO

            for index in range(len(serverData.networkData.channels)):
                channelID = serverData.networkData.channels[index]
                name = self.state.channelDefinitions[index].name
                self.log.info("%(channelName)s <---> Channel #%(channelId)d", {"channelName": name, "channelId": channelID})
                self.state.channelMap[channelID] = name

            # Replace the server's public key with our own key so we can decrypt the incoming client random
            cert = serverData.securityData.serverCertificate
            if cert:
                cert = ProprietaryCertificate(
                    cert.signatureAlgorithmID,
                    cert.keyAlgorithmID,
                    cert.publicKeyType,
                    self.state.rc4RSAKey,
                    cert.signatureType,
                    cert.signature,
                    cert.padding
                )

            # FIPS is not implemented so avoid using that
            security = ServerSecurityData(
                serverData.securityData.encryptionMethod if serverData.securityData.encryptionMethod != EncryptionMethod.ENCRYPTION_FIPS else EncryptionMethod.ENCRYPTION_128BIT,
                serverData.securityData.encryptionLevel if serverData.securityData.encryptionLevel != EncryptionLevel.ENCRYPTION_LEVEL_FIPS else EncryptionLevel.ENCRYPTION_LEVEL_HIGH,
                serverData.securityData.serverRandom,
                cert
            )

            # The clientRequestedProtocols field MUST be the same as the one received in the X224 Connection Request
            serverData.coreData.clientRequestedProtocols = self.state.requestedProtocols

            modifiedServerData = ServerDataPDU(serverData.coreData, security, serverData.networkData)
            modifiedGCCPDU = GCCConferenceCreateResponsePDU(gccPDU.nodeID, gccPDU.tag, gccPDU.result, rdpParser.write(modifiedServerData))
            modifiedMCSPDU = MCSConnectResponsePDU(pdu.result, pdu.calledConnectID, pdu.domainParams, gccParser.write(modifiedGCCPDU))

            self.client.sendPDU(modifiedMCSPDU)

    def onErectDomainRequest(self, pdu: MCSErectDomainRequestPDU):
        """
        Forward an erect domain request to the server.
        :param pdu: the erect domain request
        """
        self.server.sendPDU(pdu)

    def onAttachUserRequest(self, pdu: MCSAttachUserRequestPDU):
        """
        Forward an attach user request to the server.
        :param pdu: the attach user request
        """
        self.server.sendPDU(pdu)

    def onAttachUserConfirm(self, pdu: MCSAttachUserConfirmPDU):
        """
        Forward an attach user confirm to the client.
        :param pdu: the attach user confirm
        """
        self.client.sendPDU(pdu)

    def onChannelJoinRequest(self, pdu: MCSChannelJoinRequestPDU):
        """
        Forward a channel join request to the server.
        :param pdu: the channel join request
        """
        self.server.sendPDU(pdu)

    def onChannelJoinConfirm(self, pdu: MCSChannelJoinConfirmPDU):
        """
        If the channel join was successful, build a client and a server MCS channel and call the callback.
        :param pdu: the confirmation PDU
        """

        if pdu.result == 0:
            clientChannel = MCSServerChannel(self.client, pdu.initiator, pdu.channelID)
            serverChannel = MCSClientChannel(self.server, pdu.initiator, pdu.channelID)
            self.clientChannels[pdu.channelID] = clientChannel
            self.serverChannels[pdu.channelID] = serverChannel
            self.buildChannelCallback(clientChannel, serverChannel)

        self.client.sendPDU(pdu)

    def onSendDataRequest(self, pdu: MCSSendDataRequestPDU):
        """
        Forward a send data request to a server-side channel.
        :param pdu: the send data request
        """

        self.statCounter.increment(STAT.MCS, STAT.MCS_INPUT)

        if pdu.channelID in self.serverChannels:
            self.statCounter.increment(STAT.MCS_INPUT_ + str(pdu.channelID))
            self.clientChannels[pdu.channelID].recv(pdu.payload)

    def onSendDataIndication(self, pdu: MCSSendDataIndicationPDU):
        """
        Forward a send data indication to a client-side channel.
        :param pdu: the send data indication
        """

        self.statCounter.increment(STAT.MCS, STAT.MCS_OUTPUT)

        if pdu.channelID in self.clientChannels:
            self.statCounter.increment(STAT.MCS_OUTPUT_ + str(pdu.channelID))
            self.serverChannels[pdu.channelID].recv(pdu.payload)

    def onClientDisconnectProviderUltimatum(self, pdu: MCSDisconnectProviderUltimatumPDU):
        """
        Forward a client disconnect provider ultimatum to the server.
        :param pdu: the disconnect provider ultimatum
        """
        self.server.sendPDU(pdu)

    def onServerDisconnectProviderUltimatum(self, pdu: MCSDisconnectProviderUltimatumPDU):
        """
        Forward a server disconnect provider ultimatum to the client.
        :param pdu: the disconnect provider ultimatum
        """
        self.client.sendPDU(pdu)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import Dict, List, Union

from pyrdp.core import decodeUTF16LE, Uint16LE, Uint32LE, Uint64LE, Uint8
from pyrdp.enum import DeviceRedirectionComponent, DeviceRedirectionPacketID, \
    DeviceType, FileAccessMask, FileAttributes, FileCreateDisposition, \
    FileCreateOptions, FileShareAccess, FileSystemInformationClass, \
    GeneralCapabilityVersion, MajorFunction, MinorFunction, \
    RDPDRCapabilityType, NTSTATUS
from pyrdp.logging import log
from pyrdp.parser import Parser
from pyrdp.pdu import DeviceAnnounce, DeviceCloseRequestPDU, DeviceCloseResponsePDU, DeviceCreateRequestPDU, \
    DeviceCreateResponsePDU, DeviceDirectoryControlResponsePDU, DeviceIORequestPDU, DeviceIOResponsePDU, \
    DeviceListAnnounceRequest, DeviceQueryDirectoryRequestPDU, DeviceQueryDirectoryResponsePDU, DeviceReadRequestPDU, \
    DeviceReadResponsePDU, DeviceRedirectionCapabilitiesPDU, DeviceRedirectionCapability, \
    DeviceRedirectionClientCapabilitiesPDU, DeviceRedirectionGeneralCapability, DeviceRedirectionPDU, \
    DeviceRedirectionServerCapabilitiesPDU, FileBothDirectoryInformation, FileDirectoryInformation, \
    FileFullDirectoryInformation, FileNamesInformation


class DeviceRedirectionParser(Parser):
    """
    Parser for the device redirection channel (rdpdr) packets. Some features are missing.
    """

    def __init__(self):
        super().__init__()

        self.parsers: Dict[DeviceRedirectionPacketID, callable] = {
            DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOREQUEST: self.parseDeviceIORequest,
            DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOCOMPLETION: self.parseDeviceIOResponse,
            DeviceRedirectionPacketID.PAKID_CORE_DEVICELIST_ANNOUNCE: self.parseDeviceListAnnounce,
            DeviceRedirectionPacketID.PAKID_CORE_CLIENT_CAPABILITY: self.parseClientCapabilities,
            DeviceRedirectionPacketID.PAKID_CORE_SERVER_CAPABILITY: self.parseServerCapabilities,
        }

        self.writers: Dict[DeviceRedirectionPacketID, callable] = {
            DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOREQUEST: self.writeDeviceIORequest,
            DeviceRedirectionPacketID.PAKID_CORE_DEVICE_IOCOMPLETION: self.writeDeviceIOResponse,
            DeviceRedirectionPacketID.PAKID_CORE_DEVICELIST_ANNOUNCE: self.writeDeviceListAnnounce,
            DeviceRedirectionPacketID.PAKID_CORE_CLIENT_CAPABILITY: self.writeCapabilities,
            DeviceRedirectionPacketID.PAKID_CORE_SERVER_CAPABILITY: self.writeCapabilities,
        }

        self.capabilityParsers: Dict[RDPDRCapabilityType, callable] = {
            RDPDRCapabilityType.CAP_GENERAL_TYPE: self.parseGeneralCapability,
        }

        self.capabilityWriters: Dict[RDPDRCapabilityType, callable] = {
            RDPDRCapabilityType.CAP_GENERAL_TYPE: self.writeGeneralCapability,
        }

        self.ioRequestParsers: Dict[MajorFunction, callable] = {
            MajorFunction.IRP_MJ_CREATE: self.parseDeviceCreateRequest,
            MajorFunction.IRP_MJ_READ: self.parseDeviceReadRequest,
            MajorFunction.IRP_MJ_CLOSE: self.parseDeviceCloseRequest,
            MajorFunction.IRP_MJ_DIRECTORY_CONTROL: self.parseDirectoryControlRequest,
        }

        self.ioRequestWriters: Dict[MajorFunction, callable] = {
            MajorFunction.IRP_MJ_CREATE: self.writeDeviceCreateRequest,
            MajorFunction.IRP_MJ_READ: self.writeDeviceReadRequest,
            MajorFunction.IRP_MJ_CLOSE: self.writeDeviceCloseRequest,
            MajorFunction.IRP_MJ_DIRECTORY_CONTROL: self.writeDirectoryControlRequest,
        }

        self.ioResponseParsers: Dict[MajorFunction, callable] = {
            MajorFunction.IRP_MJ_CREATE: self.parseDeviceCreateResponse,
            MajorFunction.IRP_MJ_READ: self.parseDeviceReadResponse,
            MajorFunction.IRP_MJ_CLOSE: self.parseDeviceCloseResponse,
            MajorFunction.IRP_MJ_DIRECTORY_CONTROL: self.parseDirectoryControlResponse,
        }

        self.ioResponseWriters: Dict[MajorFunction, callable] = {
            MajorFunction.IRP_MJ_CREATE: self.writeDeviceCreateResponse,
            MajorFunction.IRP_MJ_READ: self.writeDeviceReadResponse,
            MajorFunction.IRP_MJ_CLOSE: self.writeDeviceCloseResponse,
            MajorFunction.IRP_MJ_DIRECTORY_CONTROL: self.writeDirectoryControlResponse,
        }

        self.fileInformationParsers: Dict[FileSystemInformationClass, callable] = {
            FileSystemInformationClass.FileDirectoryInformation: self.parseFileDirectoryInformation,
            FileSystemInformationClass.FileFullDirectoryInformation: self.parseFileFullDirectoryInformation,
            FileSystemInformationClass.FileBothDirectoryInformation: self.parseFileBothDirectoryInformation,
            FileSystemInformationClass.FileNamesInformation: self.parseFileNamesInformation,
        }

        self.fileInformationWriters: Dict[FileSystemInformationClass, callable] = {
            FileSystemInformationClass.FileDirectoryInformation: self.writeFileDirectoryInformation,
            FileSystemInformationClass.FileFullDirectoryInformation: self.writeFileFullDirectoryInformation,
            FileSystemInformationClass.FileBothDirectoryInformation: self.writeFileBothDirectoryInformation,
            FileSystemInformationClass.FileNamesInformation: self.writeFileNamesInformation,
        }

        # Dictionary to keep track of information associated to each completion ID.
        self.majorFunctionsForParsingResponse: Dict[int, MajorFunction] = {}
        self.minorFunctionsForParsingResponse: Dict[int, MinorFunction] = {}
        self.informationClassForParsingResponse: Dict[int, FileSystemInformationClass] = {}


    def doParse(self, data: bytes) -> DeviceRedirectionPDU:
        stream = BytesIO(data)
        component = DeviceRedirectionComponent(Uint16LE.unpack(stream))
        packetID = DeviceRedirectionPacketID(Uint16LE.unpack(stream))

        if component == DeviceRedirectionComponent.RDPDR_CTYP_CORE and packetID in self.parsers.keys():
            return self.parsers[packetID](stream)
        else:
            return DeviceRedirectionPDU(component, packetID, payload=stream.read())

    def write(self, pdu: DeviceRedirectionPDU) -> bytes:
        stream = BytesIO()
        Uint16LE.pack(pdu.component, stream)
        Uint16LE.pack(pdu.packetID, stream)

        if pdu.component == DeviceRedirectionComponent.RDPDR_CTYP_CORE and pdu.packetID in self.writers.keys():
            self.writers[pdu.packetID](pdu, stream)
        else:
            stream.write(pdu.payload)

        return stream.getvalue()


    def parseDeviceListAnnounce(self, stream: BytesIO) -> DeviceListAnnounceRequest:
        deviceCount = Uint32LE.unpack(stream)
        deviceList = [self.parseDeviceAnnounce(stream) for _ in range(deviceCount)]
        return DeviceListAnnounceRequest(deviceList)

    def writeDeviceListAnnounce(self, pdu: DeviceListAnnounceRequest, stream: BytesIO):
        Uint32LE.pack(len(pdu.deviceList), stream)

        for device in pdu.deviceList:
            self.writeDeviceAnnounce(device, stream)

    def parseDeviceAnnounce(self, stream: BytesIO) -> DeviceAnnounce:
        deviceType = DeviceType(Uint32LE.unpack(stream))
        deviceID = Uint32LE.unpack(stream)
        preferredDOSName = stream.read(8)
        deviceDataLength = Uint32LE.unpack(stream)
        deviceData = stream.read(deviceDataLength)

        preferredDOSName = preferredDOSName.decode(errors = "ignore")[: 7]

        try:
            endIndex = preferredDOSName.index("\x00")
        except ValueError:
            endIndex = -1

        if endIndex >= 0:
            preferredDOSName = preferredDOSName[: endIndex]

        return DeviceAnnounce(deviceType, deviceID, preferredDOSName, deviceData)

    def writeDeviceAnnounce(self, pdu: DeviceAnnounce, stream: BytesIO):
        Uint32LE.pack(pdu.deviceType, stream)
        Uint32LE.pack(pdu.deviceID, stream)
        stream.write(pdu.preferredDOSName.encode().ljust(7, b"\x00")[: 7] + b"\x00")
        Uint32LE.pack(len(pdu.deviceData), stream)
        stream.write(pdu.deviceData)

    def parseClientCapabilities(self, stream: BytesIO) -> DeviceRedirectionClientCapabilitiesPDU:
        capabilities = self.parseCapabilities(stream)
        return DeviceRedirectionClientCapabilitiesPDU(capabilities)

    def parseServerCapabilities(self, stream: BytesIO) -> DeviceRedirectionServerCapabilitiesPDU:
        capabilities = self.parseCapabilities(stream)
        return DeviceRedirectionServerCapabilitiesPDU(capabilities)

    def parseCapabilities(self, stream: BytesIO) -> Dict[RDPDRCapabilityType, DeviceRedirectionCapability]:
        numCapabilities = Uint16LE.unpack(stream)
        stream.read(2)  # Padding

        capabilities = {}

        for _ in range(numCapabilities):
            capability = self.parseCapability(stream)
            capabilities[capability.capabilityType] = capability

        return capabilities

    def writeCapabilities(self, pdu: DeviceRedirectionCapabilitiesPDU, stream: BytesIO):
        Uint16LE.pack(len(pdu.capabilities), stream)
        stream.write(b"\x00" * 2)  # Padding

        for capability in pdu.capabilities.values():
            self.writeCapability(capability, stream)


    def parseCapability(self, stream: BytesIO) -> DeviceRedirectionCapability:
        capabilityType = RDPDRCapabilityType(Uint16LE.unpack(stream))
        capabilityLength = Uint16LE.unpack(stream)
        version = Uint32LE.unpack(stream)
        payload = stream.read(capabilityLength - 8)

        if capabilityType in self.capabilityParsers:
            return self.capabilityParsers[capabilityType](version, payload)
        else:
            return DeviceRedirectionCapability(capabilityType, version, payload)

    def writeCapability(self, capability: DeviceRedirectionCapability, stream: BytesIO):
        Uint16LE.pack(capability.capabilityType, stream)
        substream = BytesIO()

        if capability.capabilityType in self.capabilityWriters:
            self.capabilityWriters[capability.capabilityType](capability, substream)
        else:
            substream.write(capability.payload)

        Uint16LE.pack(len(substream.getvalue()) + 8, stream)
        Uint32LE.pack(capability.version, stream)
        stream.write(substream.getvalue())


    def parseGeneralCapability(self, version: int, payload: bytes) -> DeviceRedirectionGeneralCapability:
        stream = BytesIO(payload)
        osType = Uint32LE.unpack(stream)
        osVersion = Uint32LE.unpack(stream)
        protocolMajorVersion = Uint16LE.unpack(stream)
        protocolMinorVersion = Uint16LE.unpack(stream)
        ioCode1 = Uint32LE.unpack(stream)
        ioCode2 = Uint32LE.unpack(stream)
        extendedPDU = Uint32LE.unpack(stream)
        extraFlags1 = Uint32LE.unpack(stream)
        extraFlags2 = Uint32LE.unpack(stream)
        specialTypeDeviceCap = None

        if version == GeneralCapabilityVersion.GENERAL_CAPABILITY_VERSION_02:
            specialTypeDeviceCap = Uint32LE.unpack(stream)

        return DeviceRedirectionGeneralCapability(
            version,
            osType,
            osVersion,
            protocolMajorVersion,
            protocolMinorVersion,
            ioCode1,
            ioCode2,
            extendedPDU,
            extraFlags1,
            extraFlags2,
            specialTypeDeviceCap
        )

    def writeGeneralCapability(self, capability: DeviceRedirectionGeneralCapability, stream: BytesIO):
        Uint32LE.pack(capability.osType, stream)
        Uint32LE.pack(capability.osVersion, stream)
        Uint16LE.pack(capability.protocolMajorVersion, stream)
        Uint16LE.pack(capability.protocolMinorVersion, stream)
        Uint32LE.pack(capability.ioCode1, stream)
        Uint32LE.pack(capability.ioCode2, stream)
        Uint32LE.pack(capability.extendedPDU, stream)
        Uint32LE.pack(capability.extraFlags1, stream)
        Uint32LE.pack(capability.extraFlags2, stream)

        if capability.version == GeneralCapabilityVersion.GENERAL_CAPABILITY_VERSION_02:
            Uint32LE.pack(capability.specialTypeDeviceCap or 0, stream)


    def parseDeviceIORequest(self, stream: BytesIO) -> DeviceIORequestPDU:
        deviceID = Uint32LE.unpack(stream)
        fileID = Uint32LE.unpack(stream)
        completionID = Uint32LE.unpack(stream)
        majorFunction = MajorFunction(Uint32LE.unpack(stream))
        minorFunction = Uint32LE.unpack(stream)

        if majorFunction == MajorFunction.IRP_MJ_DIRECTORY_CONTROL:
            minorFunction = MinorFunction(minorFunction)

        if majorFunction in self.ioRequestParsers.keys():
            self.majorFunctionsForParsingResponse[completionID] = majorFunction
            return self.ioRequestParsers[majorFunction](deviceID, fileID, completionID, minorFunction, stream)
        else:
            return DeviceIORequestPDU(deviceID, fileID, completionID, majorFunction, minorFunction, payload=stream.read())

    def writeDeviceIORequest(self, pdu: DeviceIORequestPDU, stream: BytesIO):
        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(pdu.fileID, stream)
        Uint32LE.pack(pdu.completionID, stream)
        Uint32LE.pack(pdu.majorFunction, stream)
        Uint32LE.pack(pdu.minorFunction, stream)

        if pdu.majorFunction in self.ioRequestWriters.keys():
            # Make sure to register the PDU's major function when we write too, in case this is a forged PDU.
            self.majorFunctionsForParsingResponse[pdu.completionID] = pdu.majorFunction
            self.ioRequestWriters[pdu.majorFunction](pdu, stream)
        else:
            stream.write(pdu.payload)


    def parseDeviceIOResponse(self, stream: BytesIO) -> DeviceIOResponsePDU:
        deviceID = Uint32LE.unpack(stream)
        completionID = Uint32LE.unpack(stream)
        ioStatus = NTSTATUS(Uint32LE.unpack(stream))
        if ioStatus._name_ == "STATUS_PYRDP_FAILURE":
            log.error("Unknown NTSTATUS value: %(value)", {"value": ioStatus._value_})

        majorFunction = self.majorFunctionsForParsingResponse.pop(completionID, None)

        if majorFunction in self.ioResponseParsers:
            return self.ioResponseParsers[majorFunction](deviceID, completionID, ioStatus, stream)
        else:
            # If for some reason, we don't know this completionID, we also return a raw response PDU (because majorFunction is None).
            payload = stream.read()
            return DeviceIOResponsePDU(majorFunction, deviceID, completionID, ioStatus, payload)

    def writeDeviceIOResponse(self, pdu: DeviceIOResponsePDU, stream: BytesIO):
        Uint32LE.pack(pdu.deviceID, stream)
        Uint32LE.pack(pdu.completionID, stream)
        Uint32LE.pack(pdu.ioStatus, stream)

        if pdu.majorFunction in self.ioResponseWriters:
            self.ioResponseWriters[pdu.majorFunction](pdu, stream)
        else:
            stream.write(pdu.payload)


    def parseDeviceCreateRequest(self, deviceID: int, fileID: int, completionID: int, minorFunction: int, stream: BytesIO) -> DeviceCreateRequestPDU:
        desiredAccess = FileAccessMask(Uint32LE.unpack(stream))
        allocationSize = Uint64LE.unpack(stream)
        fileAttributes = FileAttributes(Uint32LE.unpack(stream))
        sharedAccess = FileShareAccess(Uint32LE.unpack(stream))
        createDisposition = FileCreateDisposition(Uint32LE.unpack(stream))
        createOptions = FileCreateOptions(Uint32LE.unpack(stream))
        pathLength = Uint32LE.unpack(stream)
        path = stream.read(pathLength)

        path = decodeUTF16LE(path)[: -1]

        return DeviceCreateRequestPDU(
            deviceID,
            fileID,
            completionID,
            minorFunction,
            desiredAccess,
            allocationSize,
            fileAttributes,
            sharedAccess,
            createDisposition,
            createOptions,
            path
        )

    def writeDeviceCreateRequest(self, pdu: DeviceCreateRequestPDU, stream: BytesIO):
        # A null path *does not* include the null terminator in its length
        if len(pdu.path) == 0:
            path = b''
        else:
            path = (pdu.path + "\x00").encode("utf-16le")

        Uint32LE.pack(pdu.desiredAccess, stream)
        Uint64LE.pack(pdu.allocationSize, stream)
        Uint32LE.pack(pdu.fileAttributes, stream)
        Uint32LE.pack(pdu.sharedAccess, stream)
        Uint32LE.pack(pdu.createDisposition, stream)
        Uint32LE.pack(pdu.createOptions, stream)
        Uint32LE.pack(len(path), stream)
        stream.write(path)


    def parseDeviceCreateResponse(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, stream: BytesIO) -> DeviceCreateResponsePDU:
        """
        The information field is not yet parsed (it's optional).
        This one is a bit special since we need to look at previous packet before parsing it as
        a read response, and we need the packet data.
        """
        fileID = Uint32LE.unpack(stream)
        information = stream.read(1)

        if information == b"":
            information = 0
        else:
            information = Uint8.unpack(information)

        return DeviceCreateResponsePDU(deviceID, completionID, ioStatus, fileID, information)

    def writeDeviceCreateResponse(self, pdu: DeviceCreateResponsePDU, stream: BytesIO):
        Uint32LE.pack(pdu.fileID, stream)
        Uint8.pack(pdu.information, stream)


    def parseDeviceReadRequest(self, deviceID: int, fileID: int, completionID: int, minorFunction: int, stream: BytesIO) -> DeviceReadRequestPDU:
        length = Uint32LE.unpack(stream)
        offset = Uint64LE.unpack(stream)
        stream.read(20)  # Padding

        return DeviceReadRequestPDU(deviceID, fileID, completionID, minorFunction, length, offset)

    def writeDeviceReadRequest(self, pdu: DeviceReadRequestPDU, stream: BytesIO):
        Uint32LE.pack(pdu.length, stream)
        Uint64LE.pack(pdu.offset, stream)
        stream.write(b"\x00" * 20)  # Padding


    def parseDeviceReadResponse(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, stream: BytesIO) -> DeviceReadResponsePDU:
        length = Uint32LE.unpack(stream)
        payload = stream.read(length)

        return DeviceReadResponsePDU(deviceID, completionID, ioStatus, payload)

    def writeDeviceReadResponse(self, pdu: DeviceReadResponsePDU, stream: BytesIO):
        Uint32LE.pack(len(pdu.payload), stream)
        stream.write(pdu.payload)


    def parseDeviceCloseRequest(self, deviceID: int, fileID: int, completionID: int, minorFunction: int, stream: BytesIO) -> DeviceCloseRequestPDU:
        stream.read(32)  # Padding
        return DeviceCloseRequestPDU(deviceID, fileID, completionID, minorFunction)

    def writeDeviceCloseRequest(self, _: DeviceCloseRequestPDU, stream: BytesIO):
        stream.write(b"\x00" * 32)  # Padding


    def parseDeviceCloseResponse(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, stream: BytesIO) -> DeviceCloseResponsePDU:
        # NOTE: Even if the specification disagrees, we are sometimes getting
        #       one extra byte here. It doesn't seem to matter though.
        stream.read(4)  # Padding
        return DeviceCloseResponsePDU(deviceID, completionID, ioStatus)

    def writeDeviceCloseResponse(self, _: DeviceCloseResponsePDU, stream: BytesIO):
        # NOTE: Even if the specification disagrees, we are sometimes getting
        #       one extra byte in the DviceCloseResponse. We might need to
        #       send it one day.
        #stream.write(b"\x00" * 5)  # Padding
        stream.write(b"\x00" * 4)  # Padding


    def parseDirectoryControlRequest(self, deviceID: int, fileID: int, completionID: int, minorFunction: int, stream: BytesIO) -> DeviceIORequestPDU:
        self.minorFunctionsForParsingResponse[completionID] = MinorFunction(minorFunction)

        if minorFunction == MinorFunction.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
            return DeviceIORequestPDU(deviceID, fileID, completionID, MajorFunction.IRP_MJ_DIRECTORY_CONTROL, minorFunction, stream.read())
        else:
            informationClass = FileSystemInformationClass(Uint32LE.unpack(stream))
            initialQuery = Uint8.unpack(stream)
            pathLength = Uint32LE.unpack(stream)
            stream.read(23)

            path = stream.read(pathLength)
            path = decodeUTF16LE(path)[: -1]

            self.informationClassForParsingResponse[completionID] = informationClass
            return DeviceQueryDirectoryRequestPDU(deviceID, fileID, completionID, informationClass, initialQuery, path)

    def writeDirectoryControlRequest(self, pdu: Union[DeviceIORequestPDU, DeviceQueryDirectoryRequestPDU], stream: BytesIO):
        self.minorFunctionsForParsingResponse[pdu.completionID] = pdu.minorFunction

        if pdu.minorFunction == MinorFunction.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
            stream.write(pdu.payload)
        else:
            self.informationClassForParsingResponse[pdu.completionID] = pdu.informationClass

            Uint32LE.pack(pdu.informationClass, stream)
            Uint8.pack(pdu.initialQuery, stream)

            # A null path *does not* include the null terminator in its length
            if len(pdu.path) == 0:
                stream.write(b"\x00" * 4) # 32-bit path length (empty)
                stream.write(b"\x00" * 23) # protocol required padding
            else:
                path = (pdu.path + "\x00").encode("utf-16le")
                Uint32LE.pack(len(path), stream)
                stream.write(b"\x00" * 23) # protocol required padding
                stream.write(path)

    def parseDirectoryControlResponse(self, deviceID: int, completionID: int, ioStatus: NTSTATUS, stream: BytesIO) -> DeviceIOResponsePDU:
        minorFunction = self.minorFunctionsForParsingResponse.pop(completionID, None)

        if minorFunction is None:
            return DeviceIOResponsePDU(None, deviceID, completionID, ioStatus, stream.read())
        elif minorFunction == MinorFunction.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
            return DeviceIOResponsePDU(None, deviceID, completionID, ioStatus, stream.read())

        informationClass = self.informationClassForParsingResponse.pop(completionID)

        length = Uint32LE.unpack(stream)
        responseData = stream.read(length)
        endByte = stream.read(1)

        fileInformation = self.fileInformationParsers[informationClass](responseData)

        return DeviceQueryDirectoryResponsePDU(deviceID, completionID, ioStatus, informationClass, fileInformation, endByte)

    def writeDirectoryControlResponse(self, pdu: Union[DeviceDirectoryControlResponsePDU, DeviceQueryDirectoryResponsePDU], stream: BytesIO):
        if not hasattr(pdu, "minorFunction") or pdu.minorFunction == MinorFunction.IRP_MN_NOTIFY_CHANGE_DIRECTORY:
            stream.write(pdu.payload)
            return

        substream = BytesIO()
        self.fileInformationWriters[pdu.informationClass](pdu.fileInformation, substream)

        Uint32LE.pack(len(substream.getvalue()), stream)
        stream.write(substream.getvalue())
        stream.write(pdu.endByte)


    def writeFileInformationList(self, dataList: List[bytes], stream: BytesIO):
        currentOffset = 0

        for index, data in enumerate(dataList):
            isLastObject = index == len(dataList) - 1

            length = len(data)
            lengthWithOffset = length + 4

            if not isLastObject and lengthWithOffset % 8 != 0:
                alignmentLength = 8 - lengthWithOffset % 8
            else:
                alignmentLength = 0

            totalLength = lengthWithOffset + alignmentLength
            nextOffset = currentOffset + totalLength
            currentOffset = nextOffset

            Uint32LE.pack(nextOffset if not isLastObject else 0, stream)
            stream.write(data)
            stream.write(b"\x00" * alignmentLength)


    def parseFileDirectoryInformation(self, data: bytes) -> List[FileDirectoryInformation]:
        stream = BytesIO(data)
        information: List[FileDirectoryInformation] = []

        while stream.tell() < len(data):
            nextEntryOffset = Uint32LE.unpack(stream)
            fileIndex = Uint32LE.unpack(stream)
            creationTime = Uint64LE.unpack(stream)
            lastAccessTime = Uint64LE.unpack(stream)
            lastWriteTime = Uint64LE.unpack(stream)
            lastChangeTime = Uint64LE.unpack(stream)
            endOfFilePosition = Uint64LE.unpack(stream)
            allocationSize = Uint64LE.unpack(stream)
            fileAttributes = FileAttributes(Uint32LE.unpack(stream))
            fileNameLength = Uint32LE.unpack(stream)
            fileName = stream.read(fileNameLength)

            if nextEntryOffset == 0:
                break
            elif stream.tell() % 8 != 0:
                stream.read(8 - stream.tell() % 8) # alignment

            fileName = decodeUTF16LE(fileName)

            info = FileDirectoryInformation(
                fileIndex,
                creationTime,
                lastAccessTime,
                lastWriteTime,
                lastChangeTime,
                endOfFilePosition,
                allocationSize,
                fileAttributes,
                fileName
            )

            information.append(info)

        return information


    def writeFileDirectoryInformation(self, information: List[FileDirectoryInformation], stream: BytesIO):
        dataList: List[bytes] = []

        for info in information:
            substream = BytesIO()
            fileName = info.fileName.encode("utf-16le")

            Uint32LE.pack(info.fileIndex, substream)
            Uint64LE.pack(info.creationTime, substream)
            Uint64LE.pack(info.lastAccessTime, substream)
            Uint64LE.pack(info.lastWriteTime, substream)
            Uint64LE.pack(info.lastChangeTime, substream)
            Uint64LE.pack(info.endOfFilePosition, substream)
            Uint64LE.pack(info.allocationSize, substream)
            Uint32LE.pack(info.fileAttributes, substream)
            Uint32LE.pack(len(fileName), substream)
            substream.write(fileName)

            dataList.append(substream.getvalue())

        self.writeFileInformationList(dataList, stream)


    def parseFileFullDirectoryInformation(self, data: bytes) -> List[FileFullDirectoryInformation]:
        stream = BytesIO(data)
        information: List[FileFullDirectoryInformation] = []

        while stream.tell() < len(data):
            nextEntryOffset = Uint32LE.unpack(stream)
            fileIndex = Uint32LE.unpack(stream)
            creationTime = Uint64LE.unpack(stream)
            lastAccessTime = Uint64LE.unpack(stream)
            lastWriteTime = Uint64LE.unpack(stream)
            lastChangeTime = Uint64LE.unpack(stream)
            endOfFilePosition = Uint64LE.unpack(stream)
            allocationSize = Uint64LE.unpack(stream)
            fileAttributes = FileAttributes(Uint32LE.unpack(stream))
            fileNameLength = Uint32LE.unpack(stream)
            eaSize = Uint32LE.unpack(stream)
            fileName = stream.read(fileNameLength)

            if nextEntryOffset != 0:
                stream.read(8 - stream.tell() % 8) # alignment
                break

            fileName = decodeUTF16LE(fileName)

            info = FileFullDirectoryInformation(
                fileIndex,
                creationTime,
                lastAccessTime,
                lastWriteTime,
                lastChangeTime,
                endOfFilePosition,
                allocationSize,
                fileAttributes,
                eaSize,
                fileName
            )

            information.append(info)

        return information


    def writeFileFullDirectoryInformation(self, information: List[FileFullDirectoryInformation], stream: BytesIO):
        dataList: List[bytes] = []

        for info in information:
            substream = BytesIO()
            fileName = info.fileName.encode("utf-16le")

            Uint32LE.pack(info.fileIndex, substream)
            Uint64LE.pack(info.creationTime, substream)
            Uint64LE.pack(info.lastAccessTime, substream)
            Uint64LE.pack(info.lastWriteTime, substream)
            Uint64LE.pack(info.lastChangeTime, substream)
            Uint64LE.pack(info.endOfFilePosition, substream)
            Uint64LE.pack(info.allocationSize, substream)
            Uint32LE.pack(info.fileAttributes, substream)
            Uint32LE.pack(len(fileName), substream)
            Uint32LE.pack(info.eaSize, substream)
            substream.write(fileName)

            dataList.append(substream.getvalue())

        self.writeFileInformationList(dataList, stream)


    def parseFileBothDirectoryInformation(self, data: bytes) -> List[FileBothDirectoryInformation]:
        stream = BytesIO(data)
        information: List[FileBothDirectoryInformation] = []

        while stream.tell() < len(data):
            nextEntryOffset = Uint32LE.unpack(stream)
            fileIndex = Uint32LE.unpack(stream)
            creationTime = Uint64LE.unpack(stream)
            lastAccessTime = Uint64LE.unpack(stream)
            lastWriteTime = Uint64LE.unpack(stream)
            lastChangeTime = Uint64LE.unpack(stream)
            endOfFilePosition = Uint64LE.unpack(stream)
            allocationSize = Uint64LE.unpack(stream)
            fileAttributes = FileAttributes(Uint32LE.unpack(stream))
            fileNameLength = Uint32LE.unpack(stream)
            eaSize = Uint32LE.unpack(stream)
            shortNameLength = Uint8.unpack(stream)
            # stream.read(1) # reserved (not actually used, WTF Microsoft ????)
            shortName = stream.read(24)[: min(24, shortNameLength)]
            fileName = stream.read(fileNameLength)

            if nextEntryOffset != 0:
                stream.read(8 - stream.tell() % 8) # alignment
                break

            shortName = decodeUTF16LE(shortName)
            fileName = decodeUTF16LE(fileName)

            info = FileBothDirectoryInformation(
                fileIndex,
                creationTime,
                lastAccessTime,
                lastWriteTime,
                lastChangeTime,
                endOfFilePosition,
                allocationSize,
                fileAttributes,
                eaSize,
                shortName,
                fileName
            )

            information.append(info)

        return information


    def writeFileBothDirectoryInformation(self, information: List[FileBothDirectoryInformation], stream: BytesIO):
        dataList: List[bytes] = []

        for info in information:
            substream = BytesIO()
            fileName = info.fileName.encode("utf-16le")
            shortName = info.shortName.encode("utf-16le")

            Uint32LE.pack(info.fileIndex, substream)
            Uint64LE.pack(info.creationTime, substream)
            Uint64LE.pack(info.lastAccessTime, substream)
            Uint64LE.pack(info.lastWriteTime, substream)
            Uint64LE.pack(info.lastChangeTime, substream)
            Uint64LE.pack(info.endOfFilePosition, substream)
            Uint64LE.pack(info.allocationSize, substream)
            Uint32LE.pack(info.fileAttributes, substream)
            Uint32LE.pack(len(fileName), substream)
            Uint32LE.pack(info.eaSize, substream)
            Uint8.pack(len(shortName), substream)
            # stream.write(b"\x00") # reserved
            substream.write(shortName.ljust(24, b"\x00")[: 24])
            substream.write(fileName)

            dataList.append(substream.getvalue())

        self.writeFileInformationList(dataList, stream)


    def parseFileNamesInformation(self, data: bytes) -> List[FileNamesInformation]:
        stream = BytesIO(data)
        information: List[FileNamesInformation] = []

        while stream.tell() < len(data):
            nextEntryOffset = Uint32LE.unpack(stream)
            fileIndex = Uint32LE.unpack(stream)
            fileNameLength = Uint32LE.unpack(stream)
            fileName = stream.read(fileNameLength)

            if nextEntryOffset != 0:
                stream.read(8 - stream.tell() % 8) # alignment
                break

            fileName = decodeUTF16LE(fileName)

            info = FileNamesInformation(fileIndex, fileName)
            information.append(info)

        return information


    def writeFileNamesInformation(self, information: List[FileNamesInformation], stream: BytesIO):
        dataList: List[bytes] = []

        for info in information:
            substream = BytesIO()
            fileName = info.fileName.encode("utf-16le")

            Uint32LE.pack(info.fileIndex, substream)
            Uint32LE.pack(len(fileName), substream)
            substream.write(fileName)

            dataList.append(substream.getvalue())

        self.writeFileInformationList(dataList, stream)


    def convertWindowsTimeStamp(self, timeStamp: int) -> int:
        """
        Convert Windows time (in 100-ns) to Unix time (in ms).
        :param timeStamp: the Windows time stamp in 100-ns.
        """
        # Difference between Unix time epoch and Windows time epoch
        offset = 116444736000000000 # in 100-ns
        result = timeStamp - offset # in 100-ns
        return result // 10 # in ms



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import time
from typing import Callable


class Timer:
    """
    Simple timer class for counting elapsed time. The timer needs to be manually updated before calling getElapsedTime.
    """

    def __init__(self, timeFunc: Callable[[], float] = time.time):
        """
        :param timeFunc: the function to be used for getting the current time. Default: time.time
        """
        self.previousTime: float = None
        self.savedTime = 0.0
        self.timeFunc = timeFunc
        self.speed = 1.0

    def getElapsedTime(self) -> float:
        """
        Get the elapsed time.
        """
        return self.savedTime

    def update(self):
        """
        Update the timer. Does nothing if the timer is not running.
        """
        if self.previousTime is not None:
            currentTime = self.timeFunc()
            self.savedTime += self.speed * (currentTime - self.previousTime)
            self.previousTime = currentTime

    def start(self):
        """
        Start or resume the timer.
        """
        self.previousTime = self.timeFunc()

    def stop(self):
        """
        Stop the timer. It can still be resumed by calling start.
        """
        self.previousTime = None

    def reset(self, start = False):
        """
        Reset the timer and, if start is True, start it again.
        :param start: True if the timer should be started after resetting. Default: False
        """
        self.previousTime = None
        self.savedTime = 0.0

        if start:
            self.start()

    def setTime(self, currentTime: float):
        """
        Set the current time on the timer.
        :param currentTime: the elapsed time.
        """
        self.savedTime = currentTime

        if self.previousTime is not None:
            self.previousTime = self.timeFunc()

    def setSpeed(self, speed: float):
        """
        Set the speed of the timer.
        :param speed: speed of the timer (time multiplier).
        """
        if speed == 0.0:
            raise ValueError("Timer speed cannot be 0")

        self.speed = speed

    def isRunning(self) -> bool:
        """
        Returns True if the timer is currently running.
        """
        return self.previousTime is not None


#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import socket
from typing import Optional

from Crypto.PublicKey.RSA import RsaKey

from pyrdp.enum import ChannelOption, ConnectionDataType, RDPVersion, ServerCertificateType, EncryptionLevel
from pyrdp.enum.rdp import ClientCapabilityFlag, ColorDepth, ConnectionType, DesktopOrientation, EncryptionMethod, \
    HighColorDepth, KeyboardType, NegotiationProtocols, SupportedColorDepth
from pyrdp.pdu.pdu import PDU


class ClientCoreData:
    def __init__(self, version: RDPVersion, desktopWidth: int, desktopHeight: int, colorDepth: ColorDepth, sasSequence: int,
                 keyboardLayout: int, clientBuild: int, clientName: str, keyboardType: KeyboardType, keyboardSubType: int,
                 keyboardFunctionKey: int, imeFileName: bytes):
        self.header = ConnectionDataType.CLIENT_CORE
        self.version = version
        self.desktopWidth = desktopWidth
        self.desktopHeight = desktopHeight
        self.colorDepth = colorDepth
        self.sasSequence = sasSequence
        self.keyboardLayout = keyboardLayout
        self.clientBuild = clientBuild
        self.clientName = clientName
        self.keyboardType = keyboardType
        self.keyboardSubType = keyboardSubType
        self.keyboardFunctionKey = keyboardFunctionKey
        self.imeFileName = imeFileName
        self.postBeta2ColorDepth: ColorDepth = None
        self.clientProductId: int = None
        self.serialNumber: int = None
        self.highColorDepth: HighColorDepth = None
        self.supportedColorDepths: SupportedColorDepth = None
        self.earlyCapabilityFlags: ClientCapabilityFlag = None
        self.clientDigProductId: str = None
        self.connectionType: ConnectionType = None
        self.serverSelectedProtocol: NegotiationProtocols = None
        self.desktopPhysicalWidth: int = None
        self.desktopPhysicalHeight: int = None
        self.desktopOrientation: DesktopOrientation = None
        self.desktopScaleFactor: int = None
        self.deviceScaleFactor: int = None

    @staticmethod
    def generate(serverSelectedProtocol: NegotiationProtocols, desktopWidth: int = 800, desktopHeight: int = 600) -> 'ClientCoreData':
        """
        Generate a ClientCoreData structure with default values
        """
        version = RDPVersion.RDP5
        colorDepth = ColorDepth.RNS_UD_COLOR_8BPP
        sasSequence = 0xAA03
        keyboardLayout = 0
        clientBuild = 2600
        clientName = socket.gethostname()[: 15]
        keyboardType = KeyboardType.IBM_ENHANCED
        keyboardSubType = 0
        keyboardFunctionKey = 12
        imeFileName = b"\x00" * 64

        core = ClientCoreData(version, desktopWidth, desktopHeight, colorDepth, sasSequence, keyboardLayout, clientBuild, clientName, keyboardType, keyboardSubType, keyboardFunctionKey, imeFileName)
        core.postBeta2ColorDepth = ColorDepth.RNS_UD_COLOR_8BPP
        core.clientProductId = 1
        core.serialNumber = 0
        core.highColorDepth = HighColorDepth.HIGH_COLOR_16BPP
        core.supportedColorDepths = SupportedColorDepth.RNS_UD_16BPP_SUPPORT
        core.earlyCapabilityFlags = ClientCapabilityFlag.RNS_UD_CS_SUPPORT_ERRINFO_PDU
        core.clientDigProductId = "\x00" * 32
        core.connectionType = ConnectionType.CONNECTION_TYPE_UNKNOWN
        core.serverSelectedProtocol = serverSelectedProtocol

        return core


class ClientSecurityData:
    def __init__(self, encryptionMethods: EncryptionMethod, extEncryptionMethods: EncryptionMethod):
        self.header = ConnectionDataType.CLIENT_SECURITY
        self.encryptionMethods = encryptionMethods
        # extEncryptionMethods is used only for the French locale (https://msdn.microsoft.com/en-us/library/cc240511.aspx)
        self.extEncryptionMethods = extEncryptionMethods

    @staticmethod
    def generate(encryptionMethods: EncryptionMethod, isFrenchLocale: bool = False) -> 'ClientSecurityData':
        if isFrenchLocale:
            return ClientSecurityData(EncryptionMethod.ENCRYPTION_NONE, encryptionMethods)
        else:
            return ClientSecurityData(encryptionMethods, EncryptionMethod.ENCRYPTION_NONE)


class ClientChannelDefinition:
    def __init__(self, name: str, options: int):
        self.name = name
        self.options = options
    
    def __repr__(self):
        return "%s (0x%lx)" % (self.name, self.options)


class ClientNetworkData:
    def __init__(self, channelDefinitions: [ClientChannelDefinition]):
        self.header = ConnectionDataType.CLIENT_NETWORK
        self.channelDefinitions = channelDefinitions

    @staticmethod
    def generate(clipboard: bool = False, drive: bool = False, sound: bool = False) -> 'ClientNetworkData':
        definitions: [ClientChannelDefinition] = []

        if clipboard:
            definitions.append(ClientChannelDefinition(
                "cliprdr",
                ChannelOption.CHANNEL_OPTION_INITIALIZED
                | ChannelOption.CHANNEL_OPTION_COMPRESS_RDP
            ))

        if drive:
            definitions.append(ClientChannelDefinition(
                "rdpdr",
                ChannelOption.CHANNEL_OPTION_INITIALIZED
                | ChannelOption.CHANNEL_OPTION_COMPRESS_RDP
            ))

        if sound:
            definitions.append(ClientChannelDefinition(
                "rdpsnd",
                ChannelOption.CHANNEL_OPTION_INITIALIZED
            ))

        return ClientNetworkData(definitions)


class ClientClusterData:
    def __init__(self, flags: int, redirectedSessionID: int):
        self.header = ConnectionDataType.CLIENT_CLUSTER
        self.flags = flags
        self.redirectedSessionID = redirectedSessionID


class ClientDataPDU(PDU):
    def __init__(self, coreData: ClientCoreData, securityData: ClientSecurityData, networkData: ClientNetworkData, clusterData: Optional[ClientClusterData]):
        PDU.__init__(self)
        self.coreData = coreData
        self.securityData = securityData
        self.networkData = networkData
        self.clusterData = clusterData

    @staticmethod
    def generate(serverSelectedProtocol: NegotiationProtocols,
                 desktopWidth: int = 800, desktopHeight: int = 600,
                 encryptionMethods: EncryptionMethod = EncryptionMethod.ENCRYPTION_NONE, isFrenchLocale: bool = False,
                 clipboard: bool = False, drive: bool = False, sound: bool = False) -> 'ClientDataPDU':

        core = ClientCoreData.generate(serverSelectedProtocol, desktopWidth = desktopWidth, desktopHeight = desktopHeight)
        security = ClientSecurityData.generate(encryptionMethods = encryptionMethods, isFrenchLocale = isFrenchLocale)
        network = ClientNetworkData.generate(clipboard = clipboard, drive = drive, sound = sound)
        return ClientDataPDU(core, security, network, None)


class ServerCertificate:
    def __init__(self, certificateType: ServerCertificateType, publicKey: RsaKey, signature: bytes):
        self.type = certificateType
        self.publicKey = publicKey
        self.signature = signature


class ProprietaryCertificate(ServerCertificate):
    def __init__(self, signatureAlgorithmID: int, keyAlgorithmID: int, publicKeyType: int, publicKey: RsaKey, signatureType: int, signature: bytes, padding: bytes):
        ServerCertificate.__init__(self, ServerCertificateType.PROPRIETARY, publicKey, signature)
        self.signatureAlgorithmID = signatureAlgorithmID
        self.keyAlgorithmID = keyAlgorithmID
        self.publicKeyType = publicKeyType
        self.signatureType = signatureType
        self.padding = padding


class ServerCoreData:
    def __init__(self, version: int, clientRequestedProtocols: NegotiationProtocols, earlyCapabilityFlags: int):
        self.header = ConnectionDataType.SERVER_CORE
        self.version = version
        self.clientRequestedProtocols = clientRequestedProtocols
        self.earlyCapabilityFlags = earlyCapabilityFlags


class ServerNetworkData:
    def __init__(self, mcsChannelID: int, channels: [int]):
        self.header = ConnectionDataType.SERVER_NETWORK
        self.mcsChannelID = mcsChannelID
        self.channels = channels


class ServerSecurityData:
    def __init__(self, encryptionMethod: EncryptionMethod, encryptionLevel: EncryptionLevel, serverRandom: bytes, serverCertificate: ServerCertificate):
        self.header = ConnectionDataType.SERVER_SECURITY
        self.encryptionMethod = encryptionMethod
        self.encryptionLevel = encryptionLevel
        self.serverRandom = serverRandom
        self.serverCertificate = serverCertificate


class ServerDataPDU(PDU):
    def __init__(self, core: ServerCoreData, security: ServerSecurityData, network: ServerNetworkData):
        PDU.__init__(self)
        self.coreData = core
        self.securityData = security
        self.networkData = network





#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Common stream reading utilities.

All section numbers reference MS-RDPEGDI sections:
https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpegdi/745f2eee-d110-464c-8aca-06fc1814f6ad
"""
from io import BytesIO
from pyrdp.core.packing import Uint8, Int8, Uint16LE, Int16LE, Uint32LE


def read_encoded_uint16(s: BytesIO) -> int:
    """Read an encoded UINT16."""
    # 2.2.2.2.1.2.1.2
    b = Uint8.unpack(s)
    if b & 0x80:
        return (b & 0x7F) << 8 | Uint8.unpack(s)
    else:
        return b & 0x7F


def read_encoded_int16(s: BytesIO) -> int:
    # 2.2.2.2.1.2.1.3
    msb = Uint8.unpack(s)
    val = msb & 0x3F

    if msb & 0x80:
        lsb = Uint8.unpack(s)
        val = (val << 8) | lsb

    return -val if msb & 0x40 else val


def read_encoded_uint32(s: BytesIO) -> int:
    # 2.2.2.2.1.2.1.4
    b = Uint8.unpack(s)
    n = (b & 0xC0) >> 6
    if n == 0:
        return b & 0x3F
    elif n == 1:
        return (b & 0x3F) << 8 | Uint8.unpack(s)
    elif n == 2:
        return ((b & 0x3F) << 16 | Uint8.unpack(s) << 8 | Uint8.unpack(s))
    else:  # 3
        return ((b & 0x3F) << 24 |
                Uint8.unpack(s) << 16 |
                Uint8.unpack(s) << 8 |
                Uint8.unpack(s))


def read_color(s: BytesIO):
    """
    2.2.2.2.1.3.4.1.1 TS_COLORREF ->  rgb
    2.2.2.2.1.2.4.1   TS_COLOR_QUAD -> bgr
    """
    return Uint32LE.unpack(s) & 0x00FFFFFF


def read_utf16_str(s: BytesIO, size: int) -> [int]:
    return [Uint16LE.unpack(s) for _ in range(size)]  # Decode into str?


def read_glyph_bitmap(w: int, h: int, s: BytesIO) -> bytes:
    """Read and inflate a glyph bitmap."""

    # Glyph encoding is specified in section 2.2.2.2.1.2.6.1
    scanline = ((w + 7) // 8)
    size = scanline * h
    packed = s.read(size)
    pad = 4 - (size % 4)

    if pad < 4:  # Skip alignment padding.
        s.read(pad)

    # Convert to 1 byte per pixel format for debugging.
    # data = bytearray(w * h)
    # for y in range(h):
    #     line = y * w
    #     for x in range(w):
    #         bits = packed[scanline * y + (x // 8)]
    #         px = (bits >> (8 - (x % 8))) & 1
    #         data[line + x] = px
    # return data
    return packed


class Glyph:
    """
    TS_CACHE_GLYPH_DATA (2.2.2.2.1.2.5.1)
    """
    @staticmethod
    def parse(s: BytesIO) -> 'Glyph':
        self = Glyph()
        self.cacheIndex = Uint16LE.unpack(s)
        self.x = Uint16LE.unpack(s)
        self.y = Uint16LE.unpack(s)
        self.w = Uint16LE.unpack(s)
        self.h = Uint16LE.unpack(s)

        self.data = read_glyph_bitmap(self.w, self.h, s)

        return self


class GlyphV2:
    """
    TS_CACHE_GLYPH_DATA_REV2 (2.2.2.2.1.2.6.1)
    """
    @staticmethod
    def parse(s: BytesIO) -> Glyph:
        self = Glyph()

        self.cacheIndex = Uint8.unpack(s)

        self.x = read_encoded_int16(s)
        self.y = read_encoded_int16(s)
        self.w = read_encoded_uint16(s)
        self.h = read_encoded_uint16(s)

        self.data = read_glyph_bitmap(self.w, self.h, s)

        return self


BOUND_LEFT = 0x01
BOUND_TOP = 0x02
BOUND_RIGHT = 0x04
BOUND_BOTTOM = 0x08
BOUND_DELTA_LEFT = 0x10
BOUND_DELTA_TOP = 0x20
BOUND_DELTA_RIGHT = 0x40
BOUND_DELTA_BOTTOM = 0x80


class Bounds:
    """A bounding rectangle."""

    def __init__(self):
        self.left = 0
        self.top = 0
        self.bottom = 0
        self.right = 0

    def update(self, s: BytesIO):
        flags = Uint8.unpack(s)

        if flags & BOUND_LEFT:
            self.left = Int16LE.unpack(s)
        elif flags & BOUND_DELTA_LEFT:
            self.left += Int8.unpack(s)

        if flags & BOUND_TOP:
            self.top = Int16LE.unpack(s)
        elif flags & BOUND_DELTA_TOP:
            self.top += Int8.unpack(s)

        if flags & BOUND_RIGHT:
            self.right = Int16LE.unpack(s)
        elif flags & BOUND_DELTA_RIGHT:
            self.right += Int8.unpack(s)

        if flags & BOUND_BOTTOM:
            self.bottom = Int16LE.unpack(s)
        elif flags & BOUND_DELTA_BOTTOM:
            self.bottom += Int8.unpack(s)

    def __str__(self):
        return f'<Bounds {self.left}, {self.top}, {self.right - self.left}, {self.bottom - self.top}>'



#
# This file is part of the PyRDP project.
# Copyright (C) 2018-2023 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from PySide6.QtCore import QObject
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QTextEdit

from pyrdp.pdu import PlayerPDU
from pyrdp.enum import CapabilityType
from pyrdp.ui import QRemoteDesktop
from pyrdp.player.RenderingEventHandler import RenderingEventHandler
from pyrdp.logging import log


class PlayerEventHandler(RenderingEventHandler, QObject):
    """
    Qt Rendering Sink.

    This class handles the video pipeline by rendering to a Qt widget.
    """

    def __init__(self, viewer: QRemoteDesktop, text: QTextEdit):
        QObject.__init__(self)
        RenderingEventHandler.__init__(self, viewer)

        self.viewer = viewer
        self.text = text

    def onPDUReceived(self, pdu: PlayerPDU, isMainThread=False):
        # Ensure we are on the main thread.
        if not isMainThread:
            self.viewer.mainThreadHook.emit(lambda: self.onPDUReceived(pdu, True))
            return

        log.debug("Received %(pdu)s", {"pdu": pdu})

        # Call the parent PDU Received method.
        super().onPDUReceived(pdu)

    def onCapabilities(self, caps):
        # Set viewport's initial size.
        bmp = caps[CapabilityType.CAPSTYPE_BITMAP]
        (w, h) = (bmp.desktopWidth, bmp.desktopHeight)
        self.viewer.resize(w, h)

        super().onCapabilities(caps)

    def onMousePosition(self, x: int, y: int):
        self.viewer.setMousePosition(x, y)

    def writeText(self, text: str):
        self.text.moveCursor(QTextCursor.End)
        self.text.insertPlainText(text.rstrip("\x00"))

    def writeSeparator(self):
        self.writeText("\n--------------------\n")



#
# This file is part of the PyRDP project.
# Copyright (C) 2020-2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import CapabilityType, scancode
from pyrdp.pdu import FormatDataResponsePDU, FastPathUnicodeEvent

from pyrdp.pdu.player import PlayerPDU
from pyrdp.player.BaseEventHandler import BaseEventHandler
from pyrdp.parser import ClientInfoParser, ClientConnectionParser, ClipboardParser
from pyrdp.core import decodeUTF16LE

import json

JSON_KEY_INFO = "info"
JSON_KEY_EVENTS = "events"


class JSONEventHandler(BaseEventHandler):
    """
    Playback event handler that converts events to JSON.

    The structure is as follows:

        {
            "info": {
                "date": <timestamp>,
                "host": "HOSTNAME",
                "width": 1920,
                "height: 1080,
                "username": "USERNAME",
                "password": "PASSWORD",
                "domain": "DOMAIN",
            },

            "events": [
                {
                    "timestamp": 10000,
                    "type": "clipboard" | "key" | "mouse" | "unicode",
                    "data":  { ... EventData ... }
                }
            ]
        }

    Event data is specific to the type of event.

    clipboard:

        {
            "mime": "text" | "blob",
            "file": "filename" | null,
            "content": "utf8-text" | [0x41, ...]
        }

    key and unicode:
        {
            "press": true | false, // Whether it's a key press or release
            "key": "a", // Key name
            "mods": ["alt", "ctrl", "shift", ...] // Modifiers
        }

    mouse:
        {
            "x": 100,
            "y": 100,
            "buttons": [
                "left": true | false, // If present, whether pressed or released.
                "right": true | false,
                "middle": true | false,
            ]
        }
    """

    def __init__(self, filename: str, progress=None):
        """
        Construct an event handler that outputs to a JSON file.

        :param filename: The output file to write to.
        """

        self.json = {JSON_KEY_INFO: {}, JSON_KEY_EVENTS: []}
        self.filename = filename
        self.timestamp = None
        self.mods = set()
        self.progress = progress
        super().__init__()

    def onPDUReceived(self, pdu: PlayerPDU):
        # Keep track of the timestamp for event notation.
        self.timestamp = pdu.timestamp
        super().onPDUReceived(pdu)
        if self.progress:
            self.progress()

    def cleanup(self):
        # self.log.info("Flushing to disk: %s", self.filename)
        with open(self.filename, "w") as o:
            json.dump(self.json, o)
        self.json = None

    def onClientInfo(self, pdu: PlayerPDU):
        parser = ClientInfoParser()
        clientInfoPDU = parser.parse(pdu.payload)
        info = self.json[JSON_KEY_INFO]

        info["date"] = pdu.timestamp
        info["username"] = clientInfoPDU.username.replace("\x00", "")
        info["password"] = clientInfoPDU.password.replace("\x00", "")
        info["domain"] = clientInfoPDU.domain.replace("\x00", "")

    def onClientData(self, pdu: PlayerPDU):
        parser = ClientConnectionParser()
        clientDataPDU = parser.parse(pdu.payload)
        clientName = clientDataPDU.coreData.clientName.strip("\x00")
        self.json[JSON_KEY_INFO]["host"] = clientName

    def onClipboardData(self, pdu: PlayerPDU):
        parser = ClipboardParser()
        pdu = parser.parse(pdu.payload)

        if not isinstance(pdu, FormatDataResponsePDU):
            # TODO: Handle file PDUs.
            return

        data = decodeUTF16LE(pdu.requestedFormatData)
        self.json[JSON_KEY_EVENTS].append(
            {
                "timestamp": self.timestamp,
                "type": "clipboard",
                "data": {"mime": "text/plain", "content": data},
            }
        )

    def onMousePosition(self, x, y):
        self.mouse = (x, y)
        self.json[JSON_KEY_EVENTS].append(
            {
                "timestamp": self.timestamp,
                "type": "mouse",
                "data": {"x": x, "y": y, "buttons": []},
            }
        )

    def onMouseButton(self, buttons, pos):
        pressed = []
        if 1 in buttons:
            pressed.append({"left": buttons[1] != 0})
        if 2 in buttons:
            pressed.append({"right": buttons[2] != 0})
        if 3 in buttons:
            pressed.append({"middle": buttons[3] != 0})

        (x, y) = pos

        self.json[JSON_KEY_EVENTS].append(
            {
                "timestamp": self.timestamp,
                "type": "mouse",
                "data": {"x": x, "y": y, "buttons": pressed},
            }
        )

    def onCapabilities(self, caps):
        bmp = caps[CapabilityType.CAPSTYPE_BITMAP]
        (w, h) = (bmp.desktopWidth, bmp.desktopHeight)

        info = self.json[JSON_KEY_INFO]
        info["width"] = w
        info["height"] = h

        super().onCapabilities(caps)

    def onUnicode(self, event: FastPathUnicodeEvent):
        self.json[JSON_KEY_EVENTS].append(
            {
                "timestamp": event.timestamp,
                "type": "unicode",
                "data": {"press": not event.released, "key": event.text, "mods": []},
            }
        )

    def onScanCode(self, scanCode: int, isReleased: bool, isExtended: bool):
        keyName = scancode.getKeyName(
            scanCode, isExtended, self.shiftPressed, self.capsLockOn
        )

        # Update the state that tracks capitalization.
        if scanCode in [0x2A, 0x36]:
            self.shiftPressed = not isReleased
        elif scanCode == 0x3A and not isReleased:
            self.capsLockOn = not self.capsLockOn

        # Keep track of active modifiers.
        if scancode.isModifier(scanCode):
            if isReleased:
                self.mods.discard(keyName)  # No-throw
            else:
                self.mods.add(keyName)

        # Add the event
        self.json[JSON_KEY_EVENTS].append(
            {
                "timestamp": self.timestamp,
                "type": "key",
                "data": {
                    "key": keyName,
                    "press": not isReleased,
                    "mods": list(self.mods),
                },
            }
        )

    def writeText(self, text):
        pass  # Don't do anything.



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from binascii import hexlify
from typing import Union

from pyrdp.core import ObservedBy
from pyrdp.enum import EncryptionMethod, SecurityFlags
from pyrdp.layer.layer import IntermediateLayer, LayerObserver
from pyrdp.logging import log
from pyrdp.parser import BasicSecurityParser, ClientInfoParser, FIPSSecurityParser, SignedSecurityParser
from pyrdp.pdu import ClientInfoPDU, PDU, SecurityExchangePDU, SecurityPDU
from pyrdp.security import RC4Crypter, RC4CrypterProxy


class SecurityObserver(LayerObserver):
    def onSecurityExchangeReceived(self, pdu: SecurityExchangePDU):
        """
        Called when a Security Exchange PDU is received.
        """
        pass

    def onClientInfoReceived(self, data: bytes):
        """
        Called when client info data is received.
        """
        pass

    def onLicensingDataReceived(self, data: bytes):
        """
        Called when licensing data is received.
        """
        pass


@ObservedBy(SecurityObserver)
class SecurityLayer(IntermediateLayer):
    """
    Layer for security related traffic.
    """

    def __init__(self, parser: BasicSecurityParser):
        """
        :param parser: the parser to use for security traffic.
        """
        super().__init__(parser)
        self.clientInfoParser = ClientInfoParser()

    @staticmethod
    def create(encryptionMethod: EncryptionMethod, crypter: Union[RC4Crypter, RC4CrypterProxy]) -> 'SecurityLayer':
        """
        Create a security layer using the chosen encryption method and crypter.
        """
        if encryptionMethod in [EncryptionMethod.ENCRYPTION_40BIT, EncryptionMethod.ENCRYPTION_56BIT, EncryptionMethod.ENCRYPTION_128BIT]:
            parser = SignedSecurityParser(crypter)
            return SecurityLayer(parser)
        elif encryptionMethod == EncryptionMethod.ENCRYPTION_FIPS:
            parser = FIPSSecurityParser(crypter)
            return SecurityLayer(parser)

    def recv(self, data: bytes):
        pdu: SecurityPDU = self.mainParser.parse(data)

        try:
            self.dispatchPDU(pdu)
        except KeyboardInterrupt:
            raise
        except Exception:
            if isinstance(pdu, SecurityExchangePDU):
                log.error("Exception occurred when receiving Security Exchange. Data: %(securityExchangeData)s",
                          {"securityExchangeData": hexlify(data)})
            raise

    def dispatchPDU(self, pdu: SecurityPDU):
        """
        Send the PDU to the proper object depending on its type.
        :param pdu: the pdu.
        """
        if pdu.header & SecurityFlags.SEC_EXCHANGE_PKT != 0:
            if self.observer:
                self.observer.onSecurityExchangeReceived(pdu)
        elif pdu.header & SecurityFlags.SEC_INFO_PKT != 0:
            if self.observer:
                self.observer.onClientInfoReceived(pdu.payload)
        elif pdu.header & SecurityFlags.SEC_LICENSE_PKT != 0:
            if self.observer:
                self.observer.onLicensingDataReceived(pdu.payload)
        else:
            self.pduReceived(pdu)

    def sendBytes(self, data: bytes, header = 0):
        pdu = SecurityPDU(header, data)
        self.sendPDU(pdu)

    def sendSecurityExchange(self, clientRandom: bytes):
        """
        Send a security exchange PDU through the layer.
        :param clientRandom: the client random data.
        """
        pdu = SecurityExchangePDU(SecurityFlags.SEC_EXCHANGE_PKT, clientRandom + b"\x00" * 8)
        data = self.mainParser.writeSecurityExchange(pdu)
        self.previous.sendBytes(data)

    def sendClientInfo(self, pdu: ClientInfoPDU):
        """
        Send a client info PDU.
        """
        data = self.clientInfoParser.write(pdu)
        pdu = SecurityPDU(SecurityFlags.SEC_INFO_PKT, data)
        self.sendPDU(pdu)

    def sendLicensing(self, data: bytes):
        """
        Send raw licensing data.
        """
        pdu = SecurityPDU(SecurityFlags.SEC_LICENSE_PKT, data)
        self.sendPDU(pdu)

    def shouldForward(self, pdu: PDU) -> bool:
        return True


class TLSSecurityLayer(SecurityLayer):
    """
    Security layer used when the connection uses TLS.
    If securityHeadExpected is True, then the layer expects to receive a basic security header.
    Otherwise, the layer just forwards all the data it receives to the next layer.
    """

    def __init__(self, parser = BasicSecurityParser()):
        super().__init__(parser)
        self.securityHeaderExpected = False

    def recv(self, data: bytes):
        # Licensing happens in the security layer
        licensingBytes = b"\x80\x00"
        if not self.securityHeaderExpected and data[0:2] != licensingBytes:
            if self.next is not None:
                self.next.recv(data)
        else:
            SecurityLayer.recv(self, data)

    def sendBytes(self, data: bytes, header = 0):
        if not self.securityHeaderExpected:
            self.previous.sendBytes(data)
        else:
            SecurityLayer.sendBytes(self, data, header)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.enum import X224PDUType
from pyrdp.pdu.pdu import PDU


class X224PDU(PDU):
    """
    X.224 (T.125) PDU base class.
    """

    def __init__(self, header: X224PDUType, payload: bytes):
        """
        :param header: The PDU type.
        """

        super().__init__(payload)
        self.header = header


class X224ConnectionRequestPDU(X224PDU):

    def __init__(self, credit: int, destination: int, source: int, options: int, payload: bytes):
        super().__init__(X224PDUType.X224_TPDU_CONNECTION_REQUEST, payload)
        self.credit = credit
        self.destination = destination
        self.source = source
        self.options = options


class X224ConnectionConfirmPDU(X224PDU):

    def __init__(self, credit: int, destination: int, source: int, options: int, payload: bytes):
        super().__init__(X224PDUType.X224_TPDU_CONNECTION_CONFIRM, payload)
        self.credit = credit
        self.destination = destination
        self.source = source
        self.options = options


class X224DisconnectRequestPDU(X224PDU):

    def __init__(self, destination: int, source: int, reason: int, payload: bytes):
        super().__init__(X224PDUType.X224_TPDU_DISCONNECT_REQUEST, payload)
        self.destination = destination
        self.source = source
        self.reason = reason


class X224DataPDU(X224PDU):

    def __init__(self, roa: bool, eot: bool, payload: bytes):
        """
        :param roa: request of acknowledgement (this is False unless agreed upon during connection).
        :param eot: end of transmission (True if this is the last packet in a sequence).
        :param payload: the data bytes.
        """
        super().__init__(X224PDUType.X224_TPDU_DATA, payload)
        self.roa = roa
        self.eot = eot


class X224ErrorPDU(X224PDU):

    def __init__(self, destination: int, cause: int, payload: bytes = b""):
        super().__init__(X224PDUType.X224_TPDU_ERROR, payload)
        self.destination = destination
        self.cause = cause



#
# This file is part of the PyRDP project.
# Copyright (C) 2018, 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import List

from pyrdp.core import Uint32LE
from pyrdp.enum import VirtualChannelPDUFlag
from pyrdp.parser.parser import Parser
from pyrdp.pdu import VirtualChannelPDU


class VirtualChannelParser(Parser):
    """
    Parser class for VirtualChannel PDUs.
    """

    MAX_CHUNK_SIZE = 1600  # https://msdn.microsoft.com/en-us/library/cc240548.aspx

    def doParse(self, data: bytes) -> VirtualChannelPDU:
        stream = BytesIO(data)
        length = Uint32LE.unpack(stream)
        flags = Uint32LE.unpack(stream)
        payload = stream.read(length)
        return VirtualChannelPDU(flags, payload)

    def write(self, pdu: VirtualChannelPDU) -> List[bytes]:
        rawPacketList = []
        payloadStream = BytesIO(pdu.payload)
        lengthRemaining = len(pdu.payload)

        while lengthRemaining > 0:
            chunkSize = min(lengthRemaining, self.MAX_CHUNK_SIZE)
            chunk = payloadStream.read(chunkSize)
            flags = pdu.flags & ~(VirtualChannelPDUFlag.CHANNEL_FLAG_FIRST | VirtualChannelPDUFlag.CHANNEL_FLAG_LAST)

            if len(rawPacketList) == 0:
                # Means it's the first packet.
                flags |= VirtualChannelPDUFlag.CHANNEL_FLAG_FIRST

            if lengthRemaining <= self.MAX_CHUNK_SIZE:
                # Means it's the last packet.
                flags |= VirtualChannelPDUFlag.CHANNEL_FLAG_LAST

            outputStream = BytesIO()
            Uint32LE.pack(len(pdu.payload), outputStream)
            Uint32LE.pack(flags, outputStream)
            outputStream.write(chunk)

            rawPacketList.append(outputStream.getvalue())
            lengthRemaining -= chunkSize

        return rawPacketList



#
# This file is part of the PyRDP project.
# Copyright (C) 2019, 2021 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import logging


class SessionLogger(logging.LoggerAdapter):
    """
    Logger adapter that adds a sessionID variable to log messages. Also supports creating child loggers.
    """

    def __init__(self, logger: logging.Logger, sessionID: str):
        """
        Create a new logger adapter with a sessionID variable.
        :param logger: logger to wrap.
        :param sessionID: session ID value.
        """
        super().__init__(logger, {"sessionID": sessionID})
        self.sessionID = sessionID

    def createChild(self, childName: str, sessionID: str = None) -> 'SessionLogger':
        """
        Create a child logger wrapped in a SessionLogger.
        :param childName: logger name to append to the current logger's name.
        :param sessionID: the session ID for the child, or None to keep the same session ID.
        :return: a SessionLogger wrapping a logger that inherits the current logger.
        """
        if sessionID is None:
            sessionID = self.extra["sessionID"]

        logger = logging.getLogger(f"{self.name}.{childName}")
        sessionLogger = SessionLogger(logger, sessionID)

        # passdown clientIp if present: useful in all JSON log messages
        if 'clientIp' in self.extra:
            sessionLogger.extra['clientIp'] = self.extra['clientIp']

        return sessionLogger


#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from pyrdp.pdu.pdu import PDU


class BitmapUpdateData(PDU):
    """
    https://msdn.microsoft.com/en-us/library/cc240612.aspx
    """

    def __init__(self, destLeft: int, destTop: int, destRight: int, destBottom: int, width: int, heigth: int,
                 bitsPerPixel: int, flags: int, bitmapData: bytes):
        super().__init__()
        self.destLeft = destLeft
        self.destTop = destTop
        self.destRight = destRight
        self.destBottom = destBottom
        self.width = width
        self.heigth = heigth
        self.bitsPerPixel = bitsPerPixel
        self.flags = flags
        self.bitmapData = bitmapData


#
# This file is part of the PyRDP project.
# Copyright (C) 2020 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

"""
Constants, state and parsing primitives for Secondary Drawing Orders.
"""
from io import BytesIO

from pyrdp.core.packing import Uint8, Uint16LE, Uint32LE
from pyrdp.enum.orders import Secondary
from pyrdp.enum.rdp import GeneralExtraFlag, GlyphSupport
from .common import read_color, read_utf16_str, read_encoded_uint16, read_encoded_uint32, Glyph, GlyphV2

CBR2_BPP = [0, 0, 0, 8, 16, 24, 32]
BPP_CBR2 = [0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0,
            0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0]

CBR23_BPP = [0, 0, 0, 8, 16, 24, 32]
BPP_CBR23 = [0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0,
             0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0]

BMF_BPP = [0, 1, 0, 8, 16, 24, 32, 0]
BPP_BMF = [0, 1, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0,
           0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0]

CBR2_HEIGHT_SAME_AS_WIDTH = 0x01
CBR2_PERSISTENT_KEY_PRESENT = 0x02
CBR2_NO_BITMAP_COMPRESSION_HDR = 0x08
CBR2_DO_NOT_CACHE = 0x10

BITMAP_CACHE_WAITING_LIST_INDEX = 0x7FFF
CG_GLYPH_UNICODE_PRESENT = 0x100

CACHED_BRUSH = 0x80


def inflate_brush(packed: bytes) -> bytes:
    """Convert a packed 8 pixels-per-byte brush to 1 byte per pixel."""
    assert len(packed) == 8
    return bytes([(b >> n) & 1 for b in packed for n in range(8)])


def decompress_brush(s: BytesIO, bpp: int) -> bytes:
    """
    Decompress brush data.

    The brush data is encoded in reverse order
    """
    bitmap = s.read(16)

    paletteBpp = (bpp + 1) // 8
    palette = s.read(paletteBpp*4)
    brush = bytes(paletteBpp * 64)  # 8x8 = 64 pixels

    i = 0
    for y in range(8):
        y = 8 - y
        for x in range(8):
            if x % 4 == 0:
                pixel = bitmap[i]
                i += 1

            # Encoded as 2-bit per pixel: 00112233.
            color = (pixel >> (3 - (x % 4)) * 2) & 0x3

            # Copy `paletteBpp` bytes into the brush.
            src = color * paletteBpp
            dst = (y * 8 + x) * paletteBpp
            brush[dst:dst+paletteBpp] = palette[src:src+paletteBpp]

    return brush


class CacheBitmap:
    """
    Common type between all cached bitmaps.
    """

    def __init__(self, rev: int):
        self.rev = rev
        self.bpp = 0
        self.cacheId = 0
        self.cacheIndex = 0
        self.width = self.height = 0
        self.data = b''


class CacheBitmapV1(CacheBitmap):
    @staticmethod
    def parse(s: BytesIO, orderType: int, flags: int) -> CacheBitmap:
        self = CacheBitmapV1(1)

        self.cacheId = Uint8.unpack(s)

        s.read(1)  # Padding

        self.width = Uint8.unpack(s)
        self.height = Uint8.unpack(s)
        self.bpp = Uint8.unpack(s)

        bitmapLength = Uint16LE.unpack(s)
        self.cacheIndex = Uint16LE.unpack(s)

        if orderType & Secondary.CACHE_BITMAP_COMPRESSED and \
           not flags & GeneralExtraFlag.NO_BITMAP_COMPRESSION_HDR:
            self.compression = s.read(8)
            bitmapLength -= 8

        self.data = s.read(bitmapLength)

        return self

    def __str__(self):
        return (f'<CacheBitmapV1 {self.width}x{self.height}x{self.bpp}'
                f' Cache={self.cacheId}:{self.cacheIndex}>')


class CacheColorTable:
    @staticmethod
    def parse(s: BytesIO) -> 'CacheColorTable':
        self = CacheColorTable()

        self.cacheIndex = Uint8.unpack(s)
        numberColors = Uint16LE.unpack(s)

        assert numberColors == 256
        self.colors = [read_color(s) for _ in range(numberColors)]

        return self


class CacheGlyph:
    @staticmethod
    def parse(s: BytesIO, flags: int, glyph: GlyphSupport) -> 'CacheGlyph':
        """
        Parse a CACHE_GLYPH order.

        :param s: The byte stream to parse
        :param flags: The UPDATE PDU controlFlags
        :param glyph: One of Glyph or GlyphV2 classes to select the parsing strategy
        """

        self = CacheGlyph()

        if GlyphSupport.GLYPH_SUPPORT_ENCODE:
            self.cacheId = flags & 0x0F
            cGlyphs = flags >> 8
            unicodePresent = (flags >> 4 & 0x0F) & 0x01
            self.glyphs = [GlyphV2.parse(s) for _ in range(cGlyphs)]
        else:
            self.cacheId = Uint8.unpack(s)
            cGlyphs = Uint8.unpack(s)
            unicodePresent = flags & CG_GLYPH_UNICODE_PRESENT
            self.glyphs = [Glyph.parse(s) for _ in range(cGlyphs)]

        if unicodePresent and cGlyphs > 0:
            self.unicode = read_utf16_str(s, cGlyphs)

        return self


class CacheBitmapV2(CacheBitmap):
    def __init__(self):
        super().__init__(2)

        self.flags = 0
        self.key1 = self.key2 = 0
        self.height = self.width = 0

        self.cbCompFirstRowSize = 0
        self.cbCompMainBodySize = 0
        self.cbScanWidth = 0
        self.cbUncompressedSize = 0

    @staticmethod
    def parse(s: BytesIO, orderType: int, flags: int) -> CacheBitmap:
        self = CacheBitmapV2()

        self.cacheId = flags & 0x0003
        self.flags = (flags & 0xFF80) >> 7
        self.bpp = CBR2_BPP[(flags & 0x0078) >> 3]

        if self.flags & CBR2_PERSISTENT_KEY_PRESENT:
            self.key1 = Uint32LE.unpack(s)
            self.key2 = Uint32LE.unpack(s)

        if self.flags & CBR2_HEIGHT_SAME_AS_WIDTH:
            self.height = self.width = read_encoded_uint16(s)
        else:
            self.width = read_encoded_uint16(s)
            self.height = read_encoded_uint16(s)

        bitmapLength = read_encoded_uint32(s)
        self.cacheIndex = read_encoded_uint16(s)

        if self.flags & CBR2_DO_NOT_CACHE:
            self.cacheIndex = BITMAP_CACHE_WAITING_LIST_INDEX

        if orderType & Secondary.BITMAP_COMPRESSED_V2 and not \
           self.flags & CBR2_NO_BITMAP_COMPRESSION_HDR:
            # Parse compression header
            self.cbCompFirstRowSize = Uint16LE.unpack(s)
            self.cbCompMainBodySize = Uint16LE.unpack(s)
            self.cbScanWidth = Uint16LE.unpack(s)
            self.cbUncompressedSize = Uint16LE.unpack(s)

            bitmapLength = self.cbCompMainBodySize

        # Read bitmap data
        self.data = s.read(bitmapLength)

        return self

    def __str__(self):
        return (f'<CacheBitmapV2 {self.width}x{self.height}x{self.bpp}'
                f' Cache={self.cacheId}:{self.cacheIndex}>')


class CacheBitmapV3(CacheBitmap):
    @staticmethod
    def parse(s: BytesIO, flags: int) -> CacheBitmap:
        self = CacheBitmapV3(3)

        self.cacheId = flags & 0x00000003
        self.flags = (flags & 0x0000FF80) >> 7
        bitsPerPixelId = (flags & 0x00000078) >> 3

        # The spec says this should never be 0, but it is...
        self.bpp = CBR23_BPP[bitsPerPixelId]

        self.cacheIndex = Uint16LE.unpack(s)
        self.key1 = Uint32LE.unpack(s)
        self.key2 = Uint32LE.unpack(s)
        self.bpp = Uint8.unpack(s)

        compressed = Uint8.unpack(s)
        s.read(1)  # Reserved (1 bytes)

        self.codecId = Uint8.unpack(s)
        self.width = Uint16LE.unpack(s)
        self.height = Uint16LE.unpack(s)
        dataLen = Uint32LE.unpack(s)

        if compressed:  # TS_COMPRESSED_BITMAP_HEADER_EX present.
            s.read(24)  # Non-essential.

        self.data = s.read(dataLen)

        return self

    def __str__(self):
        return (f'<CacheBitmapV3 {self.width}x{self.height}x{self.bpp}'
                f' Cache={self.cacheId}:{self.cacheIndex} Codec={self.codecId}>')


class CacheBrush:
    @staticmethod
    def parse(s: BytesIO) -> 'CacheBrush':
        self = CacheBrush()

        self.cacheIndex = Uint8.unpack(s)

        iBitmapFormat = Uint8.unpack(s)
        assert iBitmapFormat >= 0 and iBitmapFormat < len(BMF_BPP)

        self.bpp = BMF_BPP[iBitmapFormat]

        cx = self.width = Uint8.unpack(s)
        cy = self.height = Uint8.unpack(s)

        Uint8.unpack(s)  # Unusued (style)
        # assert style == 0  # (2.2.2.2.1.2.7 Appendix 4) ... Not respected (:

        iBytes = Uint8.unpack(s)

        compressed = False
        if cx == 8 and cy == 8 and self.bpp == 1:  # 8x8 mono bitmap
            self.data = s.read(8)[::-1]
        else:
            if self.bpp == 8 and iBytes == 20:
                compressed = True
            elif self.bpp == 16 and iBytes == 24:
                compressed = True
            elif self.bpp == 24 and iBytes == 32:
                compressed = True

            if compressed:
                self.data = decompress_brush(s, self.bpp)
            else:
                self.data = bytes(256)  # Preallocate
                scanline = (self.bpp // 8) * 8
                for i in range(7):
                    # TODO: Verify correctness
                    o = (7-i)*scanline
                    self.data[o:o+8] = s.read(scanline)

        return self



#
# This file is part of the PyRDP project.
# Copyright (C) 2019-2022 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

import hashlib
import os
import tempfile
import typing
from logging import LoggerAdapter
from pathlib import Path


class FileMapping:
    """
    Class that maps a remote path to a local path. Used by the device redirection MITM component when it saves files
    transferred over RDP.
    """

    def __init__(self, file: typing.BinaryIO, dataPath: Path, filesystemPath: Path, filesystemDir: Path, log: LoggerAdapter):
        """
        :param file: the file handle for dataPath
        :param dataPath: path where the file is actually saved
        :param filesystemPath: the path to the replicated filesystem, which will be symlinked to dataPath
        :param log: logger
        """
        self.file = file
        self.filesystemPath = filesystemPath
        self.dataPath = dataPath
        self.filesystemDir = filesystemDir
        self.log = log
        self.written = False
        # only available once finalized (since we hash to find the name we can't know ahead of time)
        self.fileHash: str = None

    def seek(self, offset: int):
        if not self.file.closed:
            self.file.seek(offset)

    def write(self, data: bytes):
        self.file.write(data)
        self.written = True

    def _getShaHash(self):
        with open(self.dataPath, "rb") as f:
            # Note: In early 2022 we switched to sha256 for file hashes. If you
            #       want to use sha1, uncomment the next line and comment the
            #       other one below.
            #hash = hashlib.sha1()
            hash = hashlib.sha256()

            while True:
                buffer = f.read(65536)

                if len(buffer) == 0:
                    break

                hash.update(buffer)

        return hash.hexdigest()

    def finalize(self):
        if self.file.closed:
            return

        self.log.debug("Closing file %(path)s", {"path": self.dataPath})
        self.file.close()

        self.fileHash = self._getShaHash()

        # Go up one directory because files are saved to outDir / tmp while we're downloading them
        hashPath = (self.dataPath.parents[1] / self.fileHash)

        # Don't keep the file if we haven't written anything to it or it's a duplicate, otherwise rename and move to files dir
        if not self.written or hashPath.exists():
            self.dataPath.unlink()
        else:
            self.dataPath = self.dataPath.rename(hashPath)

        # Whether it's a duplicate or a new file, we need to create a link to it in the filesystem clone
        if self.written:
            self.filesystemPath.parents[0].mkdir(exist_ok=True, parents=True)

            if self.filesystemPath.exists():
                self.filesystemPath.unlink()

            # Make the symlink relative so you can move the output folder around and it will still work.
            self.filesystemPath.symlink_to(Path(os.path.relpath(hashPath, self.filesystemPath.parent)))

            self.log.info("SHA-256 '%(path)s' = '%(shasum)s'", {
                "path": str(self.filesystemPath.relative_to(self.filesystemDir)), "shasum": self.fileHash
            })

    def onDisconnection(self, reason):
        if not self.file.closed:
            self.log.info("Got disconnected with an active file mapping. Path: '%(path)s'. "
                          "Non-zero partial file transfer is kept here: '%(dataPath)s'. Closing.",
                          {"path": str(self.filesystemPath.relative_to(self.filesystemDir)), "dataPath": str(self.dataPath)})
            self.file.close()
            if not self.written:
                self.dataPath.unlink(missing_ok=True)

    @staticmethod
    def generate(remotePath: str, outDir: Path, filesystemDir: Path, log: LoggerAdapter):
        remotePath = Path(remotePath.replace("\\", "/"))
        filesystemPath = filesystemDir / remotePath.relative_to("/")

        tmpOutDir = outDir / "tmp"
        tmpOutDir.mkdir(exist_ok=True)

        handle, tmpPath = tempfile.mkstemp("", "", tmpOutDir)
        file = open(handle, "wb")

        log.info("Saving file '%(remotePath)s' to '%(localPath)s'", {
            "localPath": str(tmpPath), "remotePath": str(remotePath)
        })

        return FileMapping(file, Path(tmpPath), filesystemPath, filesystemDir, log)



#
# This file is part of the PyRDP project.
# Copyright (C) 2018 GoSecure Inc.
# Licensed under the GPLv3 or later.
#

from io import BytesIO
from typing import List

from pyrdp.core import Uint16LE
from pyrdp.enum import FastPathOutputType
from pyrdp.pdu import BitmapUpdateData


class BitmapParser:
    def parseBitmapUpdateData(self, data: bytes) -> List[BitmapUpdateData]:
        stream = BytesIO(data)

        bitmapUpdates = []
        numberRectangles = Uint16LE.unpack(stream.read(2))

        for i in range(numberRectangles):
            destLeft = Uint16LE.unpack(stream.read(2))
            destTop = Uint16LE.unpack(stream.read(2))
            destRight = Uint16LE.unpack(stream.read(2))
            destBottom = Uint16LE.unpack(stream.read(2))
            width = Uint16LE.unpack(stream.read(2))
            height = Uint16LE.unpack(stream.read(2))
            bitsPerPixel = Uint16LE.unpack(stream.read(2))
            flags = Uint16LE.unpack(stream.read(2))
            bitmapLength = Uint16LE.unpack(stream.read(2))
            bitmapData = stream.read(bitmapLength)
            bitmapUpdates.append(BitmapUpdateData(destLeft, destTop, destRight, destBottom, width, height, bitsPerPixel,
                                               flags, bitmapData))

        return bitmapUpdates

    def writeBitmapUpdateData(self, bitmaps: List[BitmapUpdateData]) -> bytes:
        stream = BytesIO()

        Uint16LE.pack(FastPathOutputType.FASTPATH_UPDATETYPE_BITMAP, stream)
        Uint16LE.pack(len(bitmaps), stream)

        for bitmap in bitmaps:
            Uint16LE.pack(bitmap.destLeft, stream)
            Uint16LE.pack(bitmap.destTop, stream)
            Uint16LE.pack(bitmap.destRight, stream)
            Uint16LE.pack(bitmap.destBottom, stream)
            Uint16LE.pack(bitmap.width, stream)
            Uint16LE.pack(bitmap.heigth, stream)
            Uint16LE.pack(bitmap.bitsPerPixel, stream)
            Uint16LE.pack(bitmap.flags, stream)
            Uint16LE.pack(len(bitmap.bitmapData), stream)
            stream.write(bitmap.bitmapData)

        return stream.getvalue()